import abc  # Abstract Base Classes


class BaseDistribution(abc.ABC):
    """
    Base class for probability distributions.
    """

    @abc.abstractmethod
    def sample(self, nb_points, device='cpu'):
        """
        Samples from the distribution

        :param nb_points: Number of points to sample (:obj:`int`)
        :param device: The desired device of returned tensor (:obj:`str`). Default is ``'cpu'``

        :return: Sampled points (:obj:`torch.tensor`)
        """
        raise NotImplemented

    @abc.abstractmethod
    def pdf(self, x):
        """
        Returns the pdf function evaluated at ``x``

        .. warning::
           The pdf may be unnormalized

        :param x: Value where the pdf should be evaluated (:obj:`torch.tensor`)

        :return: The pdf function evaluated at ``x`` (:obj:`torch.tensor`)
        """
        raise NotImplemented



import abc  # Abstract Base Classes
import torch


class BaseOptics(abc.ABC):
    """
    Base class for optical elements.
    """

    @abc.abstractmethod
    def get_ray_intersection(self, incident_rays):
        """
        Computes the times t at which the incident rays will intersect the object.

        :param incident_rays: The incident rays (:py:class:`~gradoptics.optics.ray.Rays`)

        :return: Times t (:obj:`torch.tensor`)
        """
        return NotImplemented

    @abc.abstractmethod
    @torch.no_grad()
    def intersect(self, incident_rays, t):
        """
        Returns the rays reflected or refracted by the optical element.

        :param incident_rays: The incident rays (:py:class:`~gradoptics.optics.ray.Rays`)
        :param t: The times at which the incident rays will intersect the optical element (:obj:`torch.tensor`)

        :return: (:obj:`tuple`)
                 - Reflected or refracted rays (:py:class:`~gradoptics.optics.ray.Rays`)
                 - A boolean mask that indicates which incident rays were reflected or refracted ― some rays may not be
                   reflected or refracted for example, due to total internal reflection, which is not modelled at this
                   point (:obj:`torch.tensor`)
        """
        return NotImplemented

    @abc.abstractmethod
    def plot(self, ax):
        """
        Plots the optical element on the provided axes.

        :param ax: 3d axes (:py:class:`mpl_toolkits.mplot3d.axes3d.Axes3D`)
        """
        return NotImplemented



import abc  # Abstract Base Classes
import torch

from gradoptics.optics import BaseOptics


class BoundingShape(BaseOptics):
    """
    Base class for bounding shapes.
    """

    @abc.abstractmethod
    def get_ray_intersection_(self, incident_rays):
        """
        Computes the times t_min and t_max at which the incident rays will intersect the bounding shape

        :param incident_rays: The incident rays (:py:class:`~gradoptics.optics.ray.Rays`)

        :return: (:obj:`tuple`)
                 - Times t_min (:obj:`torch.tensor`)
                 - Times t_max (:obj:`torch.tensor`)
        """
        return NotImplemented


from .ray import Ray, Rays
from .base_optics import BaseOptics
from .lens import PerfectLens, ThickLens
from .mirror import FlatMirror, CurvedMirror
from .bounding_sphere import BoundingSphere
from .sensor import Sensor
from .vector import vector3d, batch_vector, normalize_batch_vector, normalize_vector
from .window import Window
from .camera import Camera



import torch
import gradoptics as optics


def rejection_sampling(pdf, nb_point, proposal_distribution, m=None, batch_size=int(1e6), eps=1e-15, device='cpu'):
    """
    Samples from an unnormalized pdf using rejection sampling

    :param pdf: Function that evaluates the unnormalized pdf of a data point (:py:class:`~Callable`)
    :param nb_point: Number of points to sample (:obj:`int`)
    :param proposal_distribution: Proposal distribution (:py:class:`~gradoptics.optics.distributions.BaseDistribution`)
    :param m: M (:obj:`float`)
    :param batch_size: Batch size (:obj:`int`)
    :param eps: Parameter used for numerical stability (:obj:`float`). Default
                is ``'1e-15'``
    :param device: The desired device of returned tensor (:obj:`str`). Default is ``'cpu'``

    :return: Sampled points (:obj:`torch.tensor`)
    """
    accepted_data = torch.tensor([], device=device)

    while accepted_data.shape[0] < nb_point:

        # 1: sample U from [0, 1]
        u = torch.rand(batch_size, device=device)

        # 2: sample data from the proposal
        proposed_data = proposal_distribution.sample(batch_size, device=device)

        # If M is not specified, set it empirically
        if m is None:  # @Todo, share M over the different loops! (in order to keep the global extremum)
            m = (pdf(proposed_data) / proposal_distribution.pdf(proposed_data)).max()

        # 3: accept data x if u < pdf(x)/(M * pdf_proposal(x))
        condition = u <= (pdf(proposed_data) / (m * proposal_distribution.pdf(proposed_data) + eps))
        tmp_accepted_data = proposed_data[condition]
        accepted_data = torch.cat((accepted_data, tmp_accepted_data))
        # acceptance_rate = tmp_accepted_data.shape[0] / batch_size * 100

        del u
        del proposed_data
        del condition
        del tmp_accepted_data
        torch.cuda.empty_cache()
    return accepted_data[:nb_point]



import math
import torch

from gradoptics import Rays
from gradoptics.light_sources.base_light_source import BaseLightSource
from gradoptics.optics import batch_vector


class LightSourceFromDistribution(BaseLightSource):
    """
    Models a light source from a distribution. It emits rays in 4pi, with their origins sampled from the distribution.
    """

    def __init__(self, distribution, bounding_shape=None):
        """
        :param distribution: Distribution from which photons will be sampled
                             (:py:class:`~gradoptics.distributions.base_distribution.BaseDistribution`)
        :param bounding_shape: A bounding shape that bounds the light source
                               (:py:class:`~gradoptics.optics.bounding_shape.BoundingShape`). Default is ``None``

        .. note::
             A bounding shape is required if this light source is used with backward ray tracing

        """
        self.distribution = distribution
        self.bounding_shape = bounding_shape

    def sample_rays(self, nb_rays, device='cpu', sample_in_2pi=False):

        # Sample rays in 4 pi
        azimuthal_angle = torch.rand(nb_rays) * 2 * math.pi
        polar_angle = torch.arccos(1 - 2 * torch.rand(nb_rays))

        emitted_direction = batch_vector(torch.sin(polar_angle) * torch.sin(azimuthal_angle),
                                         torch.sin(polar_angle) * torch.cos(azimuthal_angle),
                                         torch.cos(polar_angle))
        # Sample in 2 pi
        if sample_in_2pi:
            emitted_direction[:, 0] = -emitted_direction[:, 0].abs()

        del azimuthal_angle
        del polar_angle
        torch.cuda.empty_cache()

        return Rays(self.distribution.sample(nb_rays, device=device).type(emitted_direction.dtype), emitted_direction,
                    device=device)

    def plot(self, ax, **kwargs):
        return self.distribution.plot(ax, **kwargs)

    def pdf(self, x):
        """
        Returns the pdf function of the distribution evaluated at ``x``

        .. warning::
           The pdf may be unnormalized

        :param x: Value where the pdf should be evaluated (:obj:`torch.tensor`)

        :return: The pdf function evaluated at ``x`` (:obj:`torch.tensor`)
        """
        return self.distribution.pdf(x)



from gradoptics.optics.ray import Rays
from gradoptics.optics import BaseOptics


class BoundingBox(BaseOptics):

    def __init__(self, width=1e-3, xc=0.2, yc=0.0, zc=0.0, roll=torch.tensor(np.pi/4), pitch=torch.tensor(0.), yaw=torch.tensor(0.)):
        super().__init__()
        self.width = width
        self.xc = xc
        self.yc = yc
        self.zc = zc
        self.roll = roll
        self.pitch = pitch
        self.yaw = yaw

    def get_ray_intersection(self, incident_rays, eps=1e-15):
        """
        @Todo
        :param incident_rays:
        :param eps:
        :return:
        """
        
        # Computes the intersection of the incident_ray with the cube
        orig_origins = incident_rays.origins
        orig_directions = incident_rays.directions

        # Instead of rotating the cube, apply the inverse rotation to the rays
        inv_rot_mat = rot_mat_3d(self.roll, self.pitch, self.yaw).T
        expanded_rot = torch.eye(4,dtype=torch.float64)
        expanded_rot[:3, :3] = inv_rot_mat
        
        translation_vec = torch.tensor([self.xc, self.yc, self.zc])

        origins = torch.matmul(expanded_rot[:-1], 
                               torch.cat((orig_origins-translation_vec, torch.ones((orig_origins.shape[0],1))), dim=1)[:, :, np.newaxis]).squeeze(dim=-1)+translation_vec
        directions = torch.matmul(expanded_rot[:-1], 
                               torch.cat((orig_directions, torch.zeros((orig_directions.shape[0],1))), dim=1)[:, :, np.newaxis]).squeeze(dim=-1)
        
        directions = directions/torch.sqrt(torch.sum(directions**2, dim=1, keepdim=True))

        # Now that we've rotated, we're axis aligned. Set up the faces of the cube
        fx0 = self.xc - self.width/2 
        fx1 = self.xc + self.width/2
        
        fy0 = self.yc - self.width/2 
        fy1 = self.yc + self.width/2
        
        fz0 = self.zc - self.width/2 
        fz1 = self.zc + self.width/2
        
        # Get the time of intersection with the planes corresponding to each face
        tx0 = ((fx0 - origins[:, 0]) / directions[:, 0])[:, None]
        tx1 = ((fx1 - origins[:, 0]) / directions[:, 0])[:, None]
        
        ty0 = ((fy0 - origins[:, 1]) / directions[:, 1])[:, None]
        ty1 = ((fy1 - origins[:, 1]) / directions[:, 1])[:, None]
        
        tz0 = ((fz0 - origins[:, 2]) / directions[:, 2])[:, None]
        tz1 = ((fz1 - origins[:, 2]) / directions[:, 2])[:, None]
        
        # And the corresponding points of intersection
        pointx0 = origins + tx0*directions
        pointx1 = origins + tx1*directions
        
        pointy0 = origins + ty0*directions
        pointy1 = origins + ty1*directions
        
        pointz0 = origins + tz0*directions
        pointz1 = origins + tz1*directions
        
        # Package everything up in a 6 face x n rays x 1 (3) tensor
        all_ts = torch.stack([tx0, tx1, 
                              ty0, ty1,
                              tz0, tz1])
            
        all_int = torch.stack([pointx0, pointx1, 
                              pointy0, pointy1, 
                              pointz0, pointz1])
        
        # Select the points that fall within the cube volume (with some tolerance)
        in_cube_mask = ((all_int[:, :, 0] >= fx0-eps) & (all_int[:, :, 0] <= fx1+eps) & 
                        (all_int[:, :, 1] >= fy0-eps) & (all_int[:, :, 1] <= fy1+eps) &
                        (all_int[:, :, 2] >= fz0-eps) & (all_int[:, :, 2] <= fz1+eps))

        all_ts[~in_cube_mask] = float('inf')
       
        # Return first intersection and corresponding (rotated) origin/direction 
        return all_ts.min(dim=0).values, origins, directions
    
    def intersect(self, incident_rays, t):
        origins = incident_rays.origins
        directions = incident_rays.directions

        # Update the origin of the incoming rays
        origins = origins + t.unsqueeze(1) * directions

        return Rays(origins,
                    directions,
                    luminosities=incident_rays.luminosities,
                    device=incident_rays.device,
                    meta=incident_rays.meta)
 
    def plot(self, ax, color='grey', alpha=0.4):
        phi = np.arange(1,10,2)*np.pi/4
        Phi, Theta = np.meshgrid(phi, phi)

        x = np.cos(Phi)*np.sin(Theta)
        y = np.sin(Phi)*np.sin(Theta)
        z = np.cos(Theta)/np.sqrt(2)
        
        ax.plot_surface(x*self.width+self.xc, 
                        y*self.width+self.yc, 
                        z*self.width+self.zc, color=color, alpha=alpha)



import torch

from gradoptics.optics.vector import normalize_vector, normalize_batch_vector


class Ray:
    """
    Models a ray characterized by an origin and a direction. The ray carries some luminosity and meta data which can
    be used for storing information, for example during surface intersections.
    """

    def __init__(self, origin, direction, luminosity=torch.tensor([1.]), meta=None, device='cpu'):
        """
        :param origin: Origin of the ray (:obj:`torch.tensor`)
        :param direction: Direction of the ray (:obj:`torch.tensor`)
        :param luminosity: Luminosity carried by the ray (:obj:`torch.tensor`)
        :param meta: Meta data carried by the ray (:obj:`dict`)
        :param device: The desired device of returned tensor (:obj:`str`)
        """

        super(Ray, self).__init__()

        self.origin = origin.to(device)
        self.direction = normalize_vector(direction).to(device)
        self.luminosity = luminosity.to(device)
        self.meta = meta if meta is not None else {}
        self.device = device

    def plot(self, ax, t):
        """
        Plots the ray at time ``t``

        :param ax: 3d axes (:py:class:`mpl_toolkits.mplot3d.axes3d.Axes3D`)
        :param t: times t (:obj:`torch.tensor`)
        """
        raise NotImplementedError("Not implemented yet.")


class Rays:
    """
    Models a batch of rays characterized by their origins and directions. The rays carry some luminosity and meta data
    which can be used for storing information, for example during surface intersections.
    """

    def __init__(self, origins, directions, luminosities=None, meta=None, device='cpu'):
        """
        :param origins: Origins of the rays (:obj:`torch.tensor`)
        :param directions: Directions of the rays (:obj:`torch.tensor`)
        :param luminosities: Luminosities carried by the rays (:obj:`torch.tensor`)
        :param meta: Meta data carried by the rays (:obj:`dict`)
        :param device: The desired device of returned tensor (:obj:`str`)
        """
        self.origins = origins.to(device)
        self.directions = normalize_batch_vector(directions).to(device)
        self.luminosities = luminosities.to(device) if luminosities is not None else torch.ones(origins.shape[0],
                                                                                                device=device)
        self.meta = meta if meta is not None else {}
        self.device = device

        assert self.origins.dtype == self.directions.dtype
        # Sanity check: dimensionality consistency
        assert self.origins.shape == self.directions.shape
        assert self.origins.shape[0] == self.luminosities.shape[0]
        for key in self.meta.keys():
            assert self.meta[key].shape[0] == self.origins.shape[0]

    def __getitem__(self, condition):
        """
        Returns the rays where the ``condition`` is true

        :param condition: Boolean tensor (:obj:`torch.tensor`)

        :return: A batch of rays where the ``condition`` is true (:py:class:`~gradoptics.optics.ray.Rays`)
        """
        meta = {}
        for key in self.meta.keys():
            meta[key] = self.meta[key][condition]

        return Rays(self.origins[condition], self.directions[condition], luminosities=self.luminosities[condition],
                    meta=meta, device=self.device)

    def __setitem__(self, condition, value):
        """
        Update the rays where the ``condition`` is true

        :param condition: Boolean tensor (:obj:`torch.tensor`)
        :param value: The new rays (:py:class:`~gradoptics.optics.ray.Rays`)
        """

        self.origins[condition] = value.origins
        self.directions[condition] = value.directions
        self.luminosities[condition] = value.luminosities

        assert len(self.meta.keys()) == len(value.meta.keys())

        for key in self.meta.keys():
            self.meta[key][condition] = value.meta[key]

    def __call__(self, t):
        """
        Returns the positions of the rays at times ``t``

        :param t: Times t (:obj:`torch.tensor`)

        :return: The positions of the rays at times t (:obj:`torch.tensor`)
        """
        return self.origins + t.unsqueeze(1) * self.directions

    def get_origin_and_direction(self):
        """
        Returns the origins and directions of the rays

        :return: The origins and directions of the batch of rays (:obj:`tuple`)
        """
        return self.origins, self.directions

    def get_size(self):
        """
        Returns the number of rays in the batch

        :return: The number of rays in the batch (:obj:`int`)
        """
        return self.origins.shape[0]

    def plot(self, ax, t, **kwargs):
        """
        Plots the rays at times ``t``

        :param ax: 3d axes (:py:class:`mpl_toolkits.mplot3d.axes3d.Axes3D`)
        :param t: times t (:obj:`torch.tensor`)
        """
        for i in range(len(t)):  # @Todo, parallelize
            point = self.origins[i] + t[i] * self.directions[i]

            ax.plot([self.origins[i, 0], point[0]], [self.origins[i, 1], point[1]], [self.origins[i, 2], point[2]],
                    **kwargs)


def empty_like(rays):
    """
    Returns an uninitialized batch of rays with the same size as ``rays``

    :param rays: the size of ``rays`` will determine the size of the new batch of rays
                 (:py:class:`~gradoptics.optics.ray.Rays`)

    :return: A new batch of rays (:py:class:`~gradoptics.optics.ray.Rays`)
    """
    origins = torch.empty_like(rays.origins)
    directions = torch.empty_like(rays.directions)
    luminosities = torch.empty_like(rays.luminosities)
    meta = {}
    for key in rays.meta.keys():
        meta[key] = torch.empty_like(rays.meta[key])

    return Rays(origins, directions, luminosities=luminosities, device=rays.device, meta=meta)


def cat(rays1, rays2):
    """
    Concatenates the two batch of rays

    :param rays1: First batch of rays (:py:class:`~gradoptics.optics.ray.Rays`)
    :param rays2: Second batch of rays (:py:class:`~gradoptics.optics.ray.Rays`)

    :return New batch of rays (:py:class:`~gradoptics.optics.ray.Rays`)
    """
    rays1.origins = torch.cat((rays1.origins, rays2.origins))
    rays1.directions = torch.cat((rays1.directions, rays2.directions))
    rays1.luminosities = torch.cat((rays1.luminosities, rays2.luminosities))

    for key in rays1.meta.keys():
        rays1.meta[key] = torch.cat((rays1.meta[key], rays2.meta[key]), dim=0)

    return rays1




class Scene:
    """
    Models a scene made up of multiple optical elements
    """

    def __init__(self, light_source):
        """
        :param light_source: A light source (:py:class:`~gradoptics.light_sources.base_light_source.BaseLightSource`)
        """
        self.light_source = light_source
        self.objects = []

    def add_object(self, obj):
        """
        Adds an optical element to the scene

        :param obj: An optical element (:py:class:`~gradoptics.optics.base_optics.BaseOptics`)
        """
        self.objects.append(obj)

    def plot(self, ax):
        """
        Plot the scene on the provided axes.

        :param ax: 3d axes (:py:class:`mpl_toolkits.mplot3d.axes3d.Axes3D`)
        """
        self.light_source.plot(ax)
        for obj in self.objects:
            obj.plot(ax)



import torch

from gradoptics.optics.base_optics import BaseOptics
from gradoptics.optics.ray import Rays
from gradoptics.optics.vector import dot_product


class Window(BaseOptics):
    """
    Models a medium with a refractive index between two parallel surfaces. Refraction is modelled using Snell's law.
    """

    def __init__(self, left_interface_x_position, right_interface_x_position, n_ext=1.000293, n_glass=1.494,
                 diameter=0.137, eps=1e-15):
        """
        :param left_interface_x_position: Position of the left interface along the x axis (:obj:`float`)
        :param right_interface_x_position: Position of the right interface along the x axis (:obj:`float`)
        :param n_ext: Refractive index of the external medium (:obj:`float`)
        :param n_glass: Refractive index of the medium (:obj:`float`)
        :param diameter: Diameter of the interfaces (:obj:`float`)
        :param eps: Parameter used for numerical stability in the different class methods (:obj:`float`). Default
                    is ``'1e-15'``
        """

        super(Window, self).__init__()
        assert right_interface_x_position > left_interface_x_position

        self.diameter = diameter
        self.right_interface_x_position = right_interface_x_position
        self.left_interface_x_position = left_interface_x_position
        self.n_ext = n_ext
        self.n_glass = n_glass
        self.eps = eps

    def _get_ray_intersection_left_interface(self, incident_rays):
        """
        Computes the times t at which the incident rays will intersect the left interface of the window
        :param incident_rays:
        :return:
        """
        origins = incident_rays.origins
        directions = incident_rays.directions
        t = (self.left_interface_x_position - origins[:, 0]) / (directions[:, 0] + self.eps)
        y = origins[:, 1] + t * directions[:, 1]
        z = origins[:, 2] + t * directions[:, 2]
        # Check that the intersection is real (t>0) and within the windows' aperture
        condition = (t > 1e-3) & ((y ** 2 + z ** 2) < (self.diameter / 2) ** 2)
        t[~condition] = float('inf')
        return t

    @torch.no_grad()
    def _get_ray_intersection_right_interface(self, incident_rays):
        """
        Computes the times t at which the incident rays will intersect the right interface of the window
        :param incident_rays:
        :return:
        """
        origins = incident_rays.origins
        directions = incident_rays.directions
        t = (self.right_interface_x_position - origins[:, 0]) / (directions[:, 0] + self.eps)
        y = origins[:, 1] + t * directions[:, 1]
        z = origins[:, 2] + t * directions[:, 2]
        # Check that the intersection is real (t>0) and within the windows' aperture
        condition = (t > 1e-3) & ((y ** 2 + z ** 2) < (self.diameter / 2) ** 2)
        t[~condition] = float('inf')
        return t

    def _get_rays_inside_window(self, incident_rays, t):

        origins = incident_rays.origins
        directions = incident_rays.directions
        origin_refracted_rays = origins + t.unsqueeze(1) * directions

        # Interaction with the first interface
        window_normal = torch.zeros(directions.shape, device=directions.device, dtype=directions.dtype)
        window_normal[:, 0] = 1
        # Check for each ray if it is coming from the left
        condition = directions[:, 0] > 0
        # Flip the normal for the rays coming from the right
        window_normal[~condition] *= -1
        mu = self.n_ext / self.n_glass
        # See https://physics.stackexchange.com/questions/435512/snells-law-in-vector-form
        tmp = 1 - mu ** 2 * (1 - (dot_product(window_normal, directions)) ** 2)
        c = dot_product(window_normal, directions)
        direction_refracted_rays = torch.sqrt(tmp).unsqueeze(1) * window_normal + mu * (
                directions - c.unsqueeze(1) * window_normal)

        return Rays(origin_refracted_rays, direction_refracted_rays, luminosities=incident_rays.luminosities,
                    device=incident_rays.device), window_normal

    def _get_t_min(self, incident_rays):
        """
        Computes the times t at which the incident rays will intersect the window
        :param incident_rays:
        :return:
        """
        t_left = self._get_ray_intersection_left_interface(incident_rays)
        t_right = self._get_ray_intersection_right_interface(incident_rays)
        # Keep the smallest t
        t = torch.empty(t_left.shape, device=t_left.device, dtype=t_left.dtype)
        condition = t_right < t_left
        t[condition] = t_right[condition]
        t[~condition] = t_left[~condition]
        t[torch.isinf(t)] = float('nan')
        return t

    def get_ray_intersection(self, incident_rays):

        # Time t at which the rays will intersect the first interface
        t1 = self._get_t_min(incident_rays)

        # Only keep the rays that have intersected the first interface
        mask = ~torch.isnan(t1)

        # Return if no rays have intersected the first interface
        if mask.sum() == 0:
            return t1

        ray_in_glass, window_normal = self._get_rays_inside_window(incident_rays[mask], t1[mask])
        # Interaction with the second interface
        t2 = self._get_t_min(ray_in_glass)
        nan_mask = torch.isnan(t2)
        t2[~nan_mask] = t1[mask][~nan_mask]  # t2 should only overwrite t1 where it is nan (rays that are lost)
        t1[mask] = t2
        return t1

    def intersect(self, incident_rays, t):

        ray_in_glass, window_normal = self._get_rays_inside_window(incident_rays, t)

        # Interaction with the second interface
        t = self._get_t_min(ray_in_glass)
        ray_in_glass = ray_in_glass[~torch.isnan(t)]
        incident_rays.origins = incident_rays.origins[~torch.isnan(t)]
        incident_rays.directions = incident_rays.directions[~torch.isnan(t)]
        window_normal = window_normal[~torch.isnan(t)]
        t = t[~torch.isnan(t)]
        origins = ray_in_glass.origins
        directions = ray_in_glass.directions
        origin_refracted_rays = origins + t.unsqueeze(1) * directions
        mu = self.n_glass / self.n_ext
        tmp = 1 - mu ** 2 * (1 - (dot_product(window_normal, directions)) ** 2)
        c = dot_product(window_normal, directions)
        direction_refracted_rays = torch.sqrt(tmp).unsqueeze(1) * window_normal + mu * (
                directions - c.unsqueeze(1) * window_normal)

        return (Rays(origin_refracted_rays, direction_refracted_rays, luminosities=incident_rays.luminosities,
                    meta=incident_rays.meta, device=incident_rays.device),
                torch.ones(origin_refracted_rays.shape[0], dtype=torch.bool, device=origin_refracted_rays.device))

    def plot(self, ax, s=0.1, color='lightblue', resolution=100):
        # @Todo, change this to plot_surface
        y = torch.arange(-self.diameter / 2, self.diameter / 2, self.diameter / resolution)
        z = torch.arange(-self.diameter / 2, self.diameter / 2, self.diameter / resolution)
        y, z = torch.meshgrid(y, z)

        y = y.reshape(-1)
        z = z.reshape(-1)
        x = torch.zeros(resolution * resolution)

        indices = (y ** 2 + z ** 2) < (self.diameter / 2) ** 2
        x_left = x[indices] + self.left_interface_x_position
        x_right = x[indices] + self.right_interface_x_position
        y = y[indices]
        z = z[indices]

        ax.scatter(x_left, y, z, s=s, c=color)
        ax.scatter(x_right, y, z, s=s, c=color)



import torch
import math

from gradoptics.distributions.base_distribution import BaseDistribution
from gradoptics.distributions.gaussian_distribution import GaussianDistribution
from gradoptics.inference.rejection_sampling import rejection_sampling


class AtomCloud(BaseDistribution):
    """
    Atom cloud with a sine-wave density modulation fringe pattern
    """

    def __init__(self, n=int(1e6), f=2, position=[0.31, 0., 0.], w0=0.002, h_bar=1.0546 * 1e-34, m=1.44 * 1e-25, x_a=0.,
                 y_a=0., z_a=0., t_final_bs=3., t_extra=0.1, port_bvz=.15, k_fringe=1 / 0.0003, a_quad=1e-12, phi=0.1,
                 phi2=math.pi / 2, proposal_distribution=GaussianDistribution(mean=0.0, std=0.0002)):
        """
        :param n: Number of atoms (:obj:`int`)
        :param position: Position of the center of the atom cloud (:obj:`list`)
        :param phi: Phase of the sine-wave density modulation fringe pattern (:obj:`float`)
        :param proposal_distribution: Proposal distribution used in rejection sampling for sampling from the
                                      unnormalized cloud distribution
                                      (:py:class:`~gradoptics.distributions.base_distribution.BaseDistribution`)
        """
        super().__init__()
        self.n = n
        self.f = f
        self.position = torch.tensor(position)
        self.w0 = w0
        self.h_bar = h_bar
        self.m = m
        self.xA = x_a
        self.yA = y_a
        self.zA = z_a
        self.tFinalBS = t_final_bs
        self.tExtra = t_extra
        self.portBvz = port_bvz
        self.kFringe = k_fringe
        self.aQuad = a_quad
        self.phi2 = phi2
        self.phi = phi

        # Define a sampler to sample from the cloud density (using rejection sampling)
        self.density_samplers = [lambda pdf, nb_point, device: rejection_sampling(pdf, nb_point, proposal_distribution,
                                                                                  m=None, device=device) for _ in
                                 range(3)]

    def marginal_cloud_density_x(self, x):
        """
        Returns the marginal pdf function along the x axis, evaluated at ``x``

        .. warning::
           The pdf is unnormalized

        :param x: Value where the pdf should be evaluated , in meters (:obj:`torch.tensor`)

        :return: The marginal pdf function evaluated at ``x`` (:obj:`torch.tensor`)
        """
        x = x.clone().type(torch.float64)
        x *= 10  # x to dm

        dnr = ((-1. * ((2. * self.m * self.w0 * self.w0) + (1j * self.tFinalBS * self.h_bar)) ** 1.) ** .5)
        nrc = 1j * ((2. / math.pi) ** (1. / 4.)) * ((self.m * self.w0) ** (1. / 2.))
        psi1_y = nrc * torch.exp(-1 * self.m * ((x - self.yA) ** 2) / (
                (4. * self.m * self.w0 * self.w0) + (2j * self.tFinalBS * self.h_bar))) / dnr
        density = torch.abs(psi1_y) ** 2
        return density

    def marginal_cloud_density_y(self, y):
        """
        Returns the marginal pdf function along the y axis, evaluated at ``y``

        .. warning::
           The pdf is unnormalized

        :param y: Value where the pdf should be evaluated , in meters (:obj:`torch.tensor`)

        :return: The marginal pdf function evaluated at ``y`` (:obj:`torch.tensor`)
        """

        y = y.clone().type(torch.float64)
        y *= 10  # m to dm
        dnr = ((-1. * ((2. * self.m * self.w0 * self.w0) + (1j * self.tFinalBS * self.h_bar)) ** 1.) ** .5)
        nrc = 1j * ((2. / math.pi) ** (1. / 4.)) * ((self.m * self.w0) ** (1. / 2.))
        psi1 = nrc * ((1. / (2 ** .5)) + (torch.exp(
            (1j * self.phi) + (1j * ((self.kFringe * y) + (self.aQuad * self.kFringe * self.kFringe * y * y)))) / (
                                                  2 ** .5)))
        psi1_z = psi1 * torch.exp(-1 * self.m * ((y - self.zA) ** 2) / (
                (4. * self.m * self.w0 * self.w0) + (2j * self.tFinalBS * self.h_bar))) / dnr
        density = torch.abs(psi1_z) ** 2
        return density

    def marginal_cloud_density_z(self, z):
        """
        Returns the marginal pdf function along the z axis, evaluated at ``z``

        .. warning::
           The pdf is unnormalized

        :param z: Value where the pdf should be evaluated , in meters (:obj:`torch.tensor`)

        :return: The marginal pdf function evaluated at ``z`` (:obj:`torch.tensor`)
        """

        z = z.clone().type(torch.float64)
        z *= 10  # x to dm

        dnr = ((-1. * ((2. * self.m * self.w0 * self.w0) + (1j * self.tFinalBS * self.h_bar)) ** 1.) ** .5)
        nrc = 1j * ((2. / math.pi) ** (1. / 4.)) * ((self.m * self.w0) ** (1. / 2.))
        psi1_x = nrc * torch.exp(-1 * self.m * ((z - self.xA) ** 2) / (
                (4. * self.m * self.w0 * self.w0) + (2j * self.tFinalBS * self.h_bar))) / dnr
        density = torch.abs(psi1_x) ** 2
        return density

    def pdf(self, x):  # @Todo, refractor. x,y,z -> x
        """
        Returns the pdf function evaluated at ``x``

        .. warning::
           The pdf is unnormalized

        :param x: Value where the pdf should be evaluated (:obj:`torch.tensor`)

        :return: The pdf function evaluated at ``x`` (:obj:`torch.tensor`)
        """
        return self.marginal_cloud_density_x(x[:, 0] - self.position[0]) * \
               self.marginal_cloud_density_y(x[:, 1] - self.position[1]) * \
               self.marginal_cloud_density_z(x[:, 2] - self.position[2])

    def sample(self, nb_points, device='cpu'):
        atoms = torch.empty((nb_points, 3))
        # Sample the cloud in the first dimension (Gaussian)
        tmp = self.density_samplers[0](self.marginal_cloud_density_x, nb_points, device)
        atoms[:, 0] = tmp
        # Sample the cloud in the second dimension (Gaussian)
        tmp = self.density_samplers[1](self.marginal_cloud_density_y, nb_points, device)
        atoms[:, 1] = tmp
        # Sample the cloud in the third dimension (Fringes)
        tmp = self.density_samplers[2](self.marginal_cloud_density_z, nb_points, device)
        atoms[:, 2] = tmp

        # Translate the cloud to its expected position
        ray_origins = atoms + self.position
        del atoms
        return ray_origins

    def plot(self, ax, **kwargs):
        """
        Plots the center of the atom cloud on the provided axes.

        :param ax: 3d axes (:py:class:`mpl_toolkits.mplot3d.axes3d.Axes3D`)
        """
        ax.scatter(self.position[0], self.position[1], self.position[2], **kwargs)



import abc
import torch
import numpy as np
import gradoptics as optics

from gradoptics.optics.base_optics import BaseOptics
from gradoptics.optics.ray import Rays
from gradoptics.optics.vector import batch_vector
from gradoptics.transforms.simple_transform import SimpleTransform


class Lens(BaseOptics, abc.ABC):
    """
    Base class for lenses.
    """

    @abc.abstractmethod
    def sample_points_on_lens(self, nb_points, device='cpu'):
        """
        Sample points uniformly on the lens

        :param nb_points: Number of points to sample (:obj:`int`)
        :param device: The desired device of returned tensor (:obj:`str`). Default is ``'cpu'``

        :return: Sampled points (:obj:`torch.tensor`)
        """
        raise NotImplemented


class PerfectLens(Lens):
    """
    Models a thin lens.
    """

    def __init__(self, f=0.05, na=0.714, position=None, m=0.15, transform=None, eps=1e-15):
        """
        :param f: Focal length (:obj:`float`)
        :param na: Inverse of the f-number (:obj:`float`)
        :param position: Position of the lens (:obj:`list`)
        :param m: Lens magnification  (:obj:`float`)
        :param transform: Transform to orient the lens (:py:class:`~gradoptics.transforms.base_transform.BaseTransform`)
        :param eps: Parameter used for numerical stability in the different class methods (:obj:`float`). Default
                    is ``'1e-15'``
        """
        super(PerfectLens, self).__init__()

        if position is None:
            position = [0., 0., 0.]

        self.f = f
        self.na = na
        self.m = m
        self.pof = f * (m + 1) / m
        self.eps = eps
        if transform is None:
            self.transform = SimpleTransform(0., 0., 0., torch.tensor(position))
        else:
            self.transform = transform

    def get_ray_intersection(self, incident_rays):
        incident_rays = self.transform.apply_inverse_transform(incident_rays)  # World space to lens space
        origins = incident_rays.origins
        directions = incident_rays.directions
        t = (- origins[:, 0]) / (directions[:, 0] + self.eps)
        y = origins[:, 1] + t * directions[:, 1]
        z = origins[:, 2] + t * directions[:, 2]

        # Check that the intersection is real (t>0) and within the lens' bounds
        condition = (t > 0) & ((y ** 2 + z ** 2) < (self.f * self.na / 2) ** 2)
        t[~condition] = float('nan')
        return t

    def intersect(self, incident_rays, t):
        incident_rays = self.transform.apply_inverse_transform(incident_rays)  # World space to lens space

        x_objects = incident_rays.origins
        d_0 = incident_rays.directions

        # If x_objects is on the left of the lens, put the plane of focus on the left as well
        pof = self.pof * torch.sign(x_objects[:, 0])
        # Put the camera on the opposite side of the lens with respect to x_objects
        camera_pos = - self.f * (1 + self.m) * torch.sign(x_objects[:, 0])

        # Intersections with the lens
        x_0 = x_objects + t.unsqueeze(1) * d_0

        # computes x_star, the intersections of the rays with the plane of focus
        t = (x_0[:, 0] - pof) / (d_0[:, 0] + self.eps)
        t = -t
        x_star = x_0 + t.unsqueeze(1) * d_0

        # Computes x_v, the intersections of the rays coming from x_star and passing trough the optical center with the
        # camera
        d = - x_star
        t = (camera_pos - x_star[:, 0]) / (d[:, 0] + self.eps)
        x_v = x_star + t.unsqueeze(1) * d
        _d_out = batch_vector(x_v[:, 0] - x_0[:, 0], x_v[:, 1] - x_0[:, 1], x_v[:, 2] - x_0[:, 2])

        return (self.transform.apply_transform(Rays(x_0, _d_out, luminosities=incident_rays.luminosities,
                                                    meta=incident_rays.meta, device=incident_rays.device)),
                torch.ones(x_0.shape[0], dtype=torch.bool, device=x_0.device))

    def sample_points_on_lens(self, nb_points, device='cpu'):

        points = torch.zeros((nb_points, 3), device=device)
        lens_radius = self.f * self.na / 2
        # Sampling uniformly on a disk.
        # See https://stats.stackexchange.com/questions/481543/generating-random-points-uniformly-on-a-disk
        r_squared = torch.rand(nb_points, device=device) * lens_radius ** 2
        theta = torch.rand(nb_points, device=device) * 2 * np.pi
        points[:, 1] = torch.sqrt(r_squared) * torch.cos(theta)
        points[:, 2] = torch.sqrt(r_squared) * torch.sin(theta)
        return self.transform.apply_transform_(points)

    def plot(self, ax, s=0.1, color='lightblue', resolution=100):
        # @Todo, change this to plot_surface
        y = torch.arange(-self.f * self.na / 2, self.f * self.na / 2, (self.f * self.na) / resolution)
        z = torch.arange(-self.f * self.na / 2, self.f * self.na / 2, (self.f * self.na) / resolution)
        y, z = torch.meshgrid(y, z)

        y = y.reshape(-1)
        z = z.reshape(-1)
        x = torch.zeros(resolution * resolution)

        indices = (y ** 2 + z ** 2) < (self.f * self.na / 2) ** 2
        x = x[indices]
        y = y[indices]
        z = z[indices]

        # Lens space to world space
        xyz = self.transform.apply_transform_(torch.cat((x.reshape(-1, 1), y.reshape(-1, 1), z.reshape(-1, 1)), dim=1))

        ax.scatter(xyz[:, 0], xyz[:, 1], xyz[:, 2], s=s, c=color)


class ThickLens(Lens):
    """
    Models a thick lens using Snell's law.
    """

    def __init__(self, n_lens, n_ext, R, d, transform):
        """
        :param n_lens: Index of refraction of the lens (:obj:`float`)
        :param n_ext: Index of refraction of the external medium (:obj:`float`)
        :param R: Radius of curvature of the lens (:obj:`float`)
        :param d: Thickness of the lens (:obj:`float`)
        :param transform: Transform to orient the lens (:py:class:`~gradoptics.transforms.base_transform.BaseTransform`)
        """
        super(ThickLens, self).__init__()
        self.n_lens = n_lens
        self.n_ext = n_ext
        self.R = R
        self.d = d
        self.transform = transform

    def get_focal_length(self):
        """
        Approximates the focal length of the lens based on its properties

        :return: Approximation of the focal length (:obj:`float`)
        """
        R1 = self.R
        R2 = - self.R
        return 1 / ((self.n_ext / self.n_lens - 1) * (1 / R1 - 1 / R2))

    def get_normal(self, points):
        """
        Return the normal of the lens at the given points. The normals are oriented towards the center of the lens

        :param points: 3d positions (:obj:`torch.tensor`)

        :return: Normals at the given points (:obj:`torch.tensor`)
        """

        # Computes the center of the spheres on which the points are
        center = torch.zeros_like(points)
        # -R or R as a function of the surface on which the points are
        center[:, 0] = - torch.sign(points[:, 0]) * self.R

        return optics.normalize_batch_vector(center - points)

    @staticmethod
    def _compute_intersection_with_sphere(incident_rays, R, sphere_center):
        """
        Computes the intersection of the rays ``incident_rays`` with a sphere of radius ``R``, and center
        ``sphere_center``

        :param incident_rays: The incident rays (:obj:`torch.tensor`)
        :param R: Radius of the sphere (:obj:`float`)
        :param sphere_center: Center of the sphere (:obj:`list`)

        :return:
        """
        origins, directions = incident_rays.get_origin_and_direction()

        # Solve quadratic equation in t (intersection between the incident_rays and the sphere)
        a = directions[:, 0] ** 2 + directions[:, 1] ** 2 + directions[:, 2] ** 2
        b = 2 * (((origins[:, 0] - sphere_center[0]) * directions[:, 0]) + (
                (origins[:, 1] - sphere_center[1]) * directions[:, 1]) + (
                         (origins[:, 2] - sphere_center[2]) * directions[:, 2]))
        c = (origins[:, 0] - sphere_center[0]) ** 2 + (origins[:, 1] - sphere_center[1]) ** 2 + (
                origins[:, 2] - sphere_center[2]) ** 2 - R ** 2

        pho = b ** 2 - 4 * a * c

        t1 = torch.zeros(origins.shape[0])
        t2 = torch.zeros(origins.shape[0])
        mask = pho > 0
        t1[~mask] = float('Inf')
        t2[~mask] = float('Inf')

        sqrt_pho = np.sqrt(pho[mask])
        t1[mask] = (-b[mask] + sqrt_pho) / (2 * a[mask])
        t2[mask] = (-b[mask] - sqrt_pho) / (2 * a[mask])

        # Remove virtual intersections
        t1[t1 < 0] = float('Inf')
        t2[t2 < 0] = float('Inf')
        return t1, t2

    def _get_ray_intersection_left_surface(self, incident_rays):
        """
        Computes the times t at which the incident rays will intersect the left surface of the lens

        :param incident_rays: The incident rays (:py:class:`~gradoptics.optics.ray.Rays`)

        :return: Times t (:obj:`torch.tensor`)
        """

        origin_sphere1 = [-self.R + self.d / 2, 0., 0.]
        t1, t2 = self._compute_intersection_with_sphere(incident_rays, self.R, origin_sphere1)
        # Only a portion of the surface makes up the lens. Remove intersection with other sections of the sphere
        t1[incident_rays(t1)[:, 0] < 0] = float('Inf')
        t2[incident_rays(t2)[:, 0] < 0] = float('Inf')
        return torch.min(t1, t2)

    def _get_ray_intersection_right_surface(self, incident_rays):
        """
        Computes the times t at which the incident rays will intersect the right surface of the lens

        :param incident_rays: The incident rays (:py:class:`~gradoptics.optics.ray.Rays`)

        :return: Times t (:obj:`torch.tensor`)
        """

        # Intersection with second surface
        origin_sphere2 = [self.R - self.d / 2, 0., 0.]
        t1, t2 = self._compute_intersection_with_sphere(incident_rays, self.R, origin_sphere2)
        # Only a portion of the surface makes up the lens. Remove intersection with other sections of the sphere
        t1[incident_rays(t1)[:, 0] > 0] = float('Inf')
        t2[incident_rays(t2)[:, 0] > 0] = float('Inf')
        return torch.min(t1, t2)

    def _get_rays_inside_lens(self, incident_rays, t):
        """
        Computes the rays refracted by the first surface

        :param incident_rays: The incident rays (:py:class:`~gradoptics.optics.ray.Rays`)
        :param t: The times at which the incident rays will intersect the optical element (:obj:`torch.tensor`)

        :return: (:obj:`tuple`)
                 - Reflected or refracted rays (:py:class:`~gradoptics.optics.ray.Rays`)
                 - A boolean mask that indicates which incident rays were refracted ― some rays may not be
                   refracted (:obj:`torch.tensor`)
        """

        directions = incident_rays.directions
        origin_refracted_rays = incident_rays(t)

        normals = self.get_normal(incident_rays(t))
        mu = self.n_ext / self.n_lens
        # See https://physics.stackexchange.com/questions/435512/snells-law-in-vector-form
        tmp = 1 - mu ** 2 * (1 - (optics.dot_product(normals, directions)) ** 2)
        c = optics.dot_product(normals, directions)

        total_internal_refraction_mask = tmp < 0
        mask = ~total_internal_refraction_mask

        direction_refracted_rays = torch.sqrt(tmp[mask]).unsqueeze(1) * normals[mask] + mu * (
                directions[mask] - c[mask].unsqueeze(1) * normals[mask])

        # Sanity check (should be removed)
        theta_1 = torch.arccos(
            directions[~total_internal_refraction_mask, 0] * normals[~total_internal_refraction_mask, 0] +
            directions[~total_internal_refraction_mask, 1] * normals[~total_internal_refraction_mask, 1] +
            directions[~total_internal_refraction_mask, 2] * normals[~total_internal_refraction_mask, 2])
        theta_2 = torch.arccos(direction_refracted_rays[:, 0] * normals[~total_internal_refraction_mask, 0] +
                               direction_refracted_rays[:, 1] * normals[~total_internal_refraction_mask, 1] +
                               direction_refracted_rays[:, 2] * normals[~total_internal_refraction_mask, 2])
        if theta_1.shape[0] > 0:
            assert ((torch.sin(theta_1) * self.n_ext - torch.sin(theta_2) * self.n_lens).abs().mean()) < 1e-5

        return (Rays(origin_refracted_rays[mask], direction_refracted_rays,
                     luminosities=incident_rays.luminosities[mask], device=incident_rays.device),
                mask)

    def get_ray_intersection(self, incident_rays):

        t_left = self._get_ray_intersection_left_surface(incident_rays)
        t_right = self._get_ray_intersection_right_surface(incident_rays)

        # Keep the smallest t
        t = torch.empty(t_left.shape, device=t_left.device, dtype=t_left.dtype)
        condition = t_right < t_left
        t[condition] = t_right[condition]
        t[~condition] = t_left[~condition]
        t[torch.isinf(t)] = float('nan')
        return t

    def intersect(self, incident_rays_, t_):

        incident_rays, mask = self._get_rays_inside_lens(incident_rays_, t_)

        interface_mask = incident_rays.origins[:, 0] < 0
        t = torch.empty(interface_mask.shape[0], device=mask.device)
        t[interface_mask] = self._get_ray_intersection_left_surface(incident_rays[interface_mask])
        t[~interface_mask] = self._get_ray_intersection_right_surface(incident_rays[~interface_mask])

        # A ray that was refracted by a surface may hit that surface again. Killing those rays
        mask_ = ~torch.isinf(t)

        incident_rays = incident_rays[mask_]
        t = t[mask_]

        # Interaction with the second surface
        directions = incident_rays.directions
        origin_refracted_rays = incident_rays(t)

        normal = -self.get_normal(incident_rays(t))
        mu = self.n_lens / self.n_ext
        # See https://physics.stackexchange.com/questions/435512/snells-law-in-vector-form
        tmp = 1 - mu ** 2 * (1 - (optics.dot_product(normal, directions)) ** 2)

        c = optics.dot_product(normal, directions)
        total_internal_refraction_mask = tmp < 0

        direction_refracted_rays = torch.sqrt(tmp).unsqueeze(1) * normal + mu * (directions - c.unsqueeze(1) * normal)

        # Sanity check (should be removed)
        theta_1 = torch.arccos(
            directions[~total_internal_refraction_mask, 0] * normal[~total_internal_refraction_mask, 0] +
            directions[~total_internal_refraction_mask, 1] * normal[~total_internal_refraction_mask, 1] +
            directions[~total_internal_refraction_mask, 2] * normal[~total_internal_refraction_mask, 2])
        theta_2 = torch.arccos(
            direction_refracted_rays[~total_internal_refraction_mask, 0] * normal[~total_internal_refraction_mask, 0] +
            direction_refracted_rays[~total_internal_refraction_mask, 1] * normal[~total_internal_refraction_mask, 1] +
            direction_refracted_rays[~total_internal_refraction_mask, 2] * normal[~total_internal_refraction_mask, 2])

        if theta_1.shape[0] > 0:
            assert ((torch.sin(theta_1) * self.n_lens - torch.sin(theta_2) * self.n_ext).abs().mean()) < 1e-5

        return Rays(origin_refracted_rays, direction_refracted_rays, luminosities=incident_rays.luminosities,
                    device=incident_rays.device), None

    def plot(self, ax, resolution=200, **kwargs):

        half_lens_height = np.sqrt(self.R ** 2 - (self.R - self.d / 2) ** 2)
        y = torch.arange(-half_lens_height, half_lens_height, 2 * half_lens_height / resolution)
        z = torch.arange(-half_lens_height, half_lens_height, 2 * half_lens_height / resolution)
        y, z = torch.meshgrid(y, z)

        y = y.reshape(-1)
        z = z.reshape(-1)
        x = np.sqrt(self.R ** 2 - y ** 2 - z ** 2)
        x = x - self.R + self.d / 2

        indices = x >= 0
        x = x[indices]
        y = y[indices]
        z = z[indices]

        # Plot the first lens surface
        xyz = self.transform.apply_transform_(
            torch.cat((x.reshape(-1, 1), y.reshape(-1, 1), z.reshape(-1, 1)), dim=1))  # Lens space to world space
        ax.scatter(xyz[:, 0], xyz[:, 1], xyz[:, 2], **kwargs)

        # Plot the second lens surface
        xyz = self.transform.apply_transform_(
            torch.cat((-x.reshape(-1, 1), y.reshape(-1, 1), z.reshape(-1, 1)), dim=1))  # Lens space to world space
        ax.scatter(xyz[:, 0], xyz[:, 1], xyz[:, 2], **kwargs)

    def sample_points_on_lens(self, nb_points, device='cpu'):
        raise NotImplementedError("Not implemented yet.")



import abc


class BaseIntegrator(abc.ABC):
    """
    Base class for integrators.
    """

    @abc.abstractmethod
    def compute_integral(self, incident_rays, pdf, t_min, t_max):
        """
        Computes the line integral between ``t_min`` and ``t_max`` of the incident rays through the density ``pdf``

        :param incident_rays: The rays for which the line integrals should be computed
                              (:py:class:`~gradoptics.optics.ray.Rays`)
        :param pdf: The pdf for which line integrals should be computed
        :param t_min: Lower integration bounds (:obj:`torch.tensor`)
        :param t_max: Higher integration bounds (:obj:`torch.tensor`)

        :return: Computed lines integrals (:obj:`torch.tensor`)
        """
        return NotImplemented



import torch
import numpy as np
import gradoptics as optics
import matplotlib.pyplot as plt
from gradoptics.optics import batch_vector
from gradoptics.optics.vector import cross_product
from gradoptics.optics.bounding_sphere import BoundingSphere
from gradoptics.transforms.look_at_transform import LookAtTransform
from gradoptics.transforms.simple_transform import SimpleTransform


def _test_ray(dim=1000):
    o = batch_vector(torch.zeros(dim), torch.zeros(dim), torch.zeros(dim))
    d = batch_vector(torch.ones(dim), torch.zeros(dim), torch.zeros(dim))
    ray = optics.Rays(o, d)

    if ((ray.origins.shape[0] == ray.directions.shape[0] == dim) and (
            ray.origins.shape[1] == ray.directions.shape[1] == 3)):
        return 0
    else:
        return -1


def _test_rejection_sampling():
    # Atom cloud
    atom_cloud = optics.AtomCloud()

    # Define a sampler to sample from the cloud density
    proposal_dist = optics.GaussianDistribution(mean=0., std=0.0002)
    x = optics.rejection_sampling(atom_cloud.marginal_cloud_density_x, int(1e6), proposal_dist, m=None)
    y = optics.rejection_sampling(atom_cloud.marginal_cloud_density_y, int(1e6), proposal_dist, m=None)
    z = optics.rejection_sampling(atom_cloud.marginal_cloud_density_z, int(1e6), proposal_dist, m=None)

    plt.hist(x.numpy(), bins=300, histtype='step', color='k', label='x')
    plt.hist(y.numpy(), bins=300, histtype='step', color='C0', label='y')
    plt.hist(z.numpy(), bins=300, histtype='step', color='b', label='z')
    plt.legend()
    plt.savefig('test_rejection_sampling.pdf')
    plt.close()

    return 0


def _test_atom_cloud(nb_atoms=int(1e4)):
    # Atom cloud
    atom_cloud = optics.LightSourceFromDistribution(optics.AtomCloud())

    rays = atom_cloud.sample_rays(nb_atoms)

    origins = rays.origins
    directions = rays.directions

    if not ((origins.shape[0] == directions.shape[0] == nb_atoms) and (origins.shape[1] == directions.shape[1] == 3)):
        return -1

    # Plot the origin of the rays (the cloud)
    fig = plt.figure(figsize=(12, 12))
    ax = fig.gca(projection='3d')
    ax.scatter(origins[:, 0], origins[:, 1], origins[:, 2], alpha=.3, s=0.5)
    ax.view_init(azim=0, elev=0)
    ax.set_zlim([origins[:, 2].mean() - 3 * origins[:, 2].std(), origins[:, 2].mean() + 3 * origins[:, 2].std()])
    ax.set_ylim([origins[:, 1].mean() - 3 * origins[:, 1].std(), origins[:, 1].mean() + 3 * origins[:, 1].std()])
    ax.set_xlim([origins[:, 0].mean() - 3 * origins[:, 0].std(), origins[:, 0].mean() + 3 * origins[:, 0].std()])
    plt.savefig('test_atom_cloud.pdf')
    plt.close()

    return 0


def _test_lens(nb_rays=50, f=0.05, m=0.15, right_of_lens=True, position=[0., 0., 0.]):
    lens = optics.PerfectLens(f=f, m=m, position=position)

    # Create rays parallel to the optical axis
    z_pos = torch.linspace(-lens.f * lens.na / 2 + 1e-5, lens.f * lens.na / 2 - 1e-5, nb_rays)
    origins = batch_vector(torch.zeros(nb_rays) + (1000 if right_of_lens else -1000),
                           torch.zeros(nb_rays) + position[1],
                           z_pos + position[2])

    directions = batch_vector(torch.ones(nb_rays) * (-1 if right_of_lens else 1),
                              torch.zeros(nb_rays),
                              torch.zeros(nb_rays))

    rays = optics.Rays(origins, directions)
    t = lens.get_ray_intersection(rays)
    refracted_rays, _ = lens.intersect(rays, t)

    # Computes the time t at which the rays will intersect with the image focal plane
    o = refracted_rays.origins
    d = refracted_rays.directions
    f_ = f if right_of_lens else -f
    t = (o[:, 0] - position[0] - f_) / d[:, 0]

    pt = torch.empty((nb_rays, 3))
    pt[:, 0] = o[:, 0] + t * d[:, 0]
    pt[:, 1] = o[:, 1] + t * d[:, 1]
    pt[:, 2] = o[:, 2] + t * d[:, 2]

    fig = plt.figure(figsize=(12, 12))
    ax = fig.gca(projection='3d')
    # Plot the incoming rays
    for i in range(o.shape[0]):
        ax.plot([o[i, 0] - directions[i, 0] / 10, o[i, 0]],
                [o[i, 1], o[i, 1]],
                [o[i, 2], o[i, 2]])
    # Plot the rays refracted by the lens
    for i in range(o.shape[0]):
        ax.plot([o[i, 0], pt[i, 0]],
                [o[i, 1], pt[i, 1]],
                [o[i, 2], pt[i, 2]])
    plt.savefig('test_lens.pdf')
    plt.close()

    # Check that all rays intersect at the same point
    pt_mean = pt.mean(dim=0)
    error = 0
    for point in pt:
        error += ((point - pt_mean) ** 2).sum().sqrt()

    if (error / nb_rays) < 5e-3:
        return 0
    else:
        return -1


def _test_lens_transform(nb_rays=50, f=0.05, m=0.15, right_of_lens=True, position=[0., 0., 0.]):
    transform = SimpleTransform(0., 45 / 180 * np.pi, 0., position)

    lens = optics.PerfectLens(f=f, m=m, position=position, transform=transform)

    # Create rays parallel to the optical axis
    z_pos = torch.linspace(-lens.f * lens.na / 2 + 1e-5, lens.f * lens.na / 2 - 1e-5, nb_rays)
    origins = batch_vector(torch.zeros(nb_rays) + (1000 if right_of_lens else -1000),
                           torch.zeros(nb_rays) + position[1],
                           z_pos + position[2])

    directions = batch_vector(torch.ones(nb_rays) * (-1 if right_of_lens else 1),
                              torch.zeros(nb_rays),
                              torch.zeros(nb_rays))

    rays = optics.Rays(origins, directions)
    t = lens.get_ray_intersection(rays)
    refracted_rays, _ = lens.intersect(rays, t)

    # Computes the time t at which the rays will intersect with the image focal plane
    o = refracted_rays.origins
    d = refracted_rays.directions
    f_ = f if right_of_lens else -f
    t = (o[:, 0] - position[0] - f_) / d[:, 0]

    pt = torch.empty((nb_rays, 3))
    pt[:, 0] = o[:, 0] + t * d[:, 0]
    pt[:, 1] = o[:, 1] + t * d[:, 1]
    pt[:, 2] = o[:, 2] + t * d[:, 2]

    fig = plt.figure(figsize=(12, 12))
    ax = fig.gca(projection='3d')
    # Plot the incoming rays
    for i in range(o.shape[0]):
        ax.plot([o[i, 0] - directions[i, 0] / 10, o[i, 0]],
                [o[i, 1], o[i, 1]],
                [o[i, 2], o[i, 2]])
    # Plot the rays refracted by the lens
    for i in range(o.shape[0]):
        ax.plot([o[i, 0], pt[i, 0]],
                [o[i, 1], pt[i, 1]],
                [o[i, 2], pt[i, 2]])
    plt.savefig('test_lens_transform.pdf')
    plt.close()

    return 0


def _test_thick_lens(nb_rays=64):

    position = torch.tensor([0., 0., 0.])
    transform = optics.simple_transform.SimpleTransform(0, 0, 0, position)
    lens = optics.ThickLens(1.5, 1., 1, 1e-1, transform)

    # Create rays parallel to the optical axis
    y_pos = torch.linspace(-.1, .1, int(np.sqrt(nb_rays)))
    z_pos = torch.linspace(-.1, .1, int(np.sqrt(nb_rays)))
    y_pos, z_pos = torch.meshgrid((y_pos, z_pos))

    origins = optics.batch_vector(torch.zeros(nb_rays) + 2, y_pos + position[1], z_pos + position[2])
    directions = optics.batch_vector(torch.ones(nb_rays) * -1, torch.zeros(nb_rays), torch.zeros(nb_rays))
    rays = optics.Rays(origins, directions)

    t = lens.get_ray_intersection(rays)
    mask = ~torch.isnan(t)
    refracted_rays, _ = lens.intersect(rays[mask], t[mask])

    fig = plt.figure(figsize=(12, 12))
    ax = fig.gca(projection='3d')
    rays[mask].plot(ax, t[mask])
    refracted_rays.plot(ax, [1.2 for _ in range(refracted_rays.origins.shape[0])], color='k', linestyle='--')
    ax.view_init(elev=30., azim=70)
    plt.savefig('test_thick_lens.pdf')
    plt.close()

    return 0


def _ray_marching_test_sensor(eps=1e-3):
    f = 0.05
    m = 0.15

    sensor = optics.Sensor(position=(-f * (1 + m), 0, 0))
    pts = sensor.sample_points_on_sensor(1000000)

    plt.hist2d(pts[:, 1].numpy(), pts[:, 2].numpy(), bins=40)
    plt.gca().set_aspect('equal')
    plt.colorbar()
    plt.savefig('test_ray_marching_sensor.pdf')
    plt.close()

    pts[:, 1] /= sensor.pixel_size[1]
    pts[:, 2] /= sensor.pixel_size[0]

    if ((pts[:, 1].max() - eps) < sensor.resolution[1] / 2) and \
            ((pts[:, 1].min() + eps) > -sensor.resolution[1] / 2) and \
            ((pts[:, 2].max() - eps) < sensor.resolution[0] / 2) and \
            ((pts[:, 2].min() + eps) > -sensor.resolution[0] / 2):
        return 0
    else:
        return -1


def _ray_marching_test_lens(eps=1e-3):
    f = 0.05
    na = 1 / 1.4

    lens = optics.PerfectLens(f=f, na=na)
    pts = lens.sample_points_on_lens(1000000)

    plt.figure()
    plt.hist2d(pts[:, 1].numpy(), pts[:, 2].numpy(), bins=40)
    plt.gca().set_aspect('equal')
    plt.colorbar()
    plt.savefig('test_ray_marching_lens.pdf')
    plt.close()

    if (pts.max() <= (f * na / 2 + eps)) and (pts.min() >= (-f * na / 2 - eps)):
        return 0
    else:
        return -1


def _test_window():
    obj_x_pos = 0.31
    left_interface_x_position = obj_x_pos + .056
    right_interface_x_position = left_interface_x_position + .05
    window = optics.Window(left_interface_x_position, right_interface_x_position)
    atom_cloud = optics.LightSourceFromDistribution(optics.AtomCloud(position=[obj_x_pos, 0., 0.]))
    nb_atoms = int(1e5)

    rays = atom_cloud.sample_rays(nb_atoms)
    t = window.get_ray_intersection(rays)
    r = rays[~torch.isnan(t)]
    t = t[~torch.isnan(t)]
    rays, _ = window.intersect(r, t)

    fig = plt.figure(figsize=(12, 12))
    ax = fig.gca(projection='3d')
    n = 100
    window.plot(ax)
    rays.plot(ax, [.1 for _ in range(n)], color='k')
    r.plot(ax, t[:n])
    plt.savefig('test_window.pdf')
    plt.close()

    # The refracted direction should be the same as the incident direction
    if ((rays.directions[:n] - r.directions[:n]).abs().sum() / n) < 1e-5:
        return 0
    else:
        return -1


def _test_cross_product():
    assert (cross_product(torch.tensor([1, 0, 0]), torch.tensor([0, 1, 0])) == torch.tensor([0, 0, 1])).all()
    assert (cross_product(torch.tensor([0, 1, 0]), torch.tensor([0, 0, 1])) == torch.tensor([1, 0, 0])).all()
    assert (cross_product(torch.tensor([0, 0, 1]), torch.tensor([1, 0, 0])) == torch.tensor([0, 1, 0])).all()
    assert (cross_product(torch.tensor([0, 1, 0]), torch.tensor([1, 0, 0])) == torch.tensor([0, 0, -1])).all()
    assert (cross_product(torch.tensor([0, 0, 1]), torch.tensor([0, 1, 0])) == torch.tensor([-1, 0, 0])).all()
    assert (cross_product(torch.tensor([1, 0, 0]), torch.tensor([0, 0, 1])) == torch.tensor([0, -1, 0])).all()
    return 0


def _test_look_at_transform():
    up = torch.tensor([0., 0., 1.])
    viewing_direction = torch.tensor([1., 0., 0.])
    pos = torch.tensor([0., 0., 0.])
    transform = LookAtTransform(viewing_direction, pos, up=up)

    assert (transform.transform == torch.tensor([[0., 0., 1., 0.],
                                                 [1., 0., 0., 0.],
                                                 [0., 1., 0., 0.],
                                                 [0., 0., 0., 1.]])).all()

    up = torch.tensor([0., 0., 1.])
    viewing_direction = torch.tensor([-1., 0., 0.])
    pos = torch.tensor([0., 0., 0.])
    transform = LookAtTransform(viewing_direction, pos, up)

    assert (transform.transform == (torch.tensor([[0., 0., -1., 0.],
                                                  [-1., 0., 0., 0.],
                                                  [0., 1., 0., 0.],
                                                  [0., 0., 0., 1.]]))).all()

    up = torch.tensor([1., 0., 0.])
    viewing_direction = torch.tensor([0., 0., 1.])
    pos = torch.tensor([0., 0., 0.])
    transform = LookAtTransform(viewing_direction, pos, up)

    assert (transform.transform == (torch.tensor([[0., 1., 0., 0.],
                                                  [-1., 0., 0., 0.],
                                                  [0., -0., 1., 0.],
                                                  [0., 0., 0., 1.]]))).all()

    up = torch.tensor([-1., 0., 0.])
    viewing_direction = torch.tensor([0., 0., 1.])
    pos = torch.tensor([0., 0., 0.])
    transform = LookAtTransform(viewing_direction, pos, up)

    assert (transform.transform == (torch.tensor([[0., -1., 0., 0.],
                                                  [1., 0., 0., 0.],
                                                  [-0., 0., 1., 0.],
                                                  [0., 0., 0., 1.]]))).all()
    return 0


def _test_grad_rays():
    def toy_intersect(rays: optics.Rays, normal=torch.tensor([.2, .2, .6])) -> optics.Rays:
        x = (rays.directions[:, 0] * normal[0]).reshape(-1, 1)
        y = (rays.directions[:, 1] * normal[1]).reshape(-1, 1)
        z = (rays.directions[:, 2] * normal[2]).reshape(-1, 1)
        new_directions = torch.cat((x, y, z), dim=1)
        new_origins = rays.origins * 2 + 1
        return optics.Rays(new_origins, new_directions)

    origins = torch.randn(1, 3, requires_grad=True)
    directions = torch.randn(1, 3, requires_grad=True)
    rays1 = optics.Rays(origins, directions)
    rays2 = toy_intersect(rays1)
    assert directions.is_leaf
    assert origins.is_leaf

    toy_loss = (rays2.origins * rays2.directions).sum()
    toy_loss.backward()
    assert origins.grad.abs().sum() > 0
    assert directions.grad.abs().sum() > 0
    return 0


def _test_bounding_sphere(nb_rays=1000):
    fig = plt.figure(figsize=(12, 12))
    ax = fig.gca(projection='3d')

    cloud_envelopes = [BoundingSphere(xc=3e-3, radii=1e-3),
                       BoundingSphere(xc=-3e-3, radii=1e-3),
                       BoundingSphere(xc=0, yc=3e-3, radii=1e-3),
                       BoundingSphere(xc=0, yc=-3e-3, radii=1e-3)]

    for cloud_envelope in cloud_envelopes:
        cloud_envelope.plot(ax)

    # Rays
    origins = torch.zeros((nb_rays, 3))
    # Sample rays in 4 pi
    azimuthal_angle = torch.rand(nb_rays) * 2 * np.pi
    polar_angle = torch.arccos(1 - 2 * torch.rand(nb_rays))
    emitted_direction = optics.batch_vector(torch.sin(polar_angle) * torch.sin(azimuthal_angle),
                                            torch.sin(polar_angle) * torch.cos(azimuthal_angle),
                                            torch.cos(polar_angle))
    rays = optics.Rays(origins, emitted_direction)

    # First intersection with the sphere
    t_s = []
    for cloud_envelope in cloud_envelopes:
        t_s.append(cloud_envelope.get_ray_intersection(rays))

    for t in t_s:
        rays.plot(ax, t)
        assert (t[~torch.isnan(t)] > 0).all()  # Check that all the times are greater than 0

    # Second intersection with the sphere
    for i, cloud_envelope in enumerate(cloud_envelopes):
        cond = ~torch.isnan(t_s[i])
        outgoing_rays, _ = cloud_envelope.intersect(rays[cond], t_s[i][cond])
        t_ = cloud_envelope.get_ray_intersection(outgoing_rays)
        outgoing_rays.plot(ax, t_, color='r')
        cond_ = torch.isnan(t_)
        assert cond_.sum() == 0  # In this setup, All the rays should make it to the second intersection
        assert torch.allclose(rays[cond].directions,
                              outgoing_rays.directions)  # The sphere should not modify the directions
        outgoing_rays, _ = cloud_envelope.intersect(outgoing_rays, t_)
        assert torch.allclose(rays[cond].directions,
                              outgoing_rays.directions)  # The sphere should not modify the directions

    ax.set_xlim([-0.004, 0.004])
    ax.set_ylim([-0.004, 0.004])
    ax.set_zlim([-0.004, 0.004])
    plt.savefig('test_bounding_sphere.pdf')
    plt.close()
    return 0


def _test_grad_mirror_wrt_incident_rays(mirror_position=torch.ones(3),
                                        ray_origins=torch.randn(2, 3, requires_grad=True)):
    mirror = optics.FlatMirror(mirror_position[0], mirror_position[1], mirror_position[2], torch.tensor([.2, .2, .6]),
                               .005)

    directions = [(mirror_position[0] - ray_origins[:, 0]).reshape(-1, 1),
                  (mirror_position[1] - ray_origins[:, 1]).reshape(-1, 1),
                  (mirror_position[2] - ray_origins[:, 2]).reshape(-1, 1)]
    directions = torch.cat(directions, dim=1)
    directions = directions.clone().detach().requires_grad_(True)
    rays = optics.Rays(ray_origins, directions)
    t = mirror.get_ray_intersection(rays)

    outgoing_rays, _ = mirror.intersect(rays, t)

    loss = (outgoing_rays.origins * outgoing_rays.directions).sum()
    loss.backward()

    assert ray_origins.is_leaf
    assert directions.is_leaf
    assert ray_origins.grad.abs().sum() > 0
    assert directions.grad.abs().sum() > 0
    return 0


def _test_grad_lens_wrt_incident_rays(lens_position=[1., 1., 1.], ray_origins=torch.randn(2, 3, requires_grad=True)):
    lens = optics.PerfectLens(position=lens_position)
    directions = [(lens_position[0] - ray_origins[:, 0]).reshape(-1, 1),
                  (lens_position[1] - ray_origins[:, 1]).reshape(-1, 1),
                  (lens_position[2] - ray_origins[:, 2]).reshape(-1, 1)]
    directions = torch.cat(directions, dim=1)
    directions = directions.clone().detach().requires_grad_(True)
    rays = optics.Rays(ray_origins, directions)
    t = lens.get_ray_intersection(rays)

    outgoing_rays, _ = lens.intersect(rays, t)

    loss = (outgoing_rays.origins * outgoing_rays.directions).sum()
    loss.backward()

    assert directions.is_leaf
    assert ray_origins.is_leaf
    assert ray_origins.grad.abs().sum() > 0
    assert directions.grad.abs().sum() > 0
    return 0


"""def _test_grad_sensor_wrt_incident_rays(sensor_position=torch.ones(3),
                                        ray_origins=torch.randn(2, 3, requires_grad=True)):
    sensor = optics.Sensor(position=sensor_position)
    directions = [(sensor_position[0] - ray_origins[:, 0]).reshape(-1, 1),
                  (sensor_position[1] - ray_origins[:, 1]).reshape(-1, 1),
                  (sensor_position[2] - ray_origins[:, 2]).reshape(-1, 1)]
    directions = torch.cat(directions, dim=1)
    directions = directions.clone().detach().requires_grad_(True)
    rays = optics.Rays(ray_origins, directions)
    t = sensor.get_ray_intersection(rays)

    hit_position, _, _ = sensor.intersect(rays, t)

    target_position = torch.randn(hit_position.shape)

    l1_loss = ((hit_position - target_position) ** 2).mean()
    l1_loss.backward()

    assert directions.is_leaf
    assert ray_origins.is_leaf
    assert ray_origins.grad.abs().sum() > 0
    assert directions.grad.abs().sum() > 0
    return 0
"""


def _test_grad_mirror_wrt_self_parameters():
    mirror_normal = torch.tensor([.2, .2, .6], requires_grad=True)
    mirror_position = torch.tensor([1., 1., 1.], requires_grad=True)
    mirror = optics.FlatMirror(mirror_position[0], mirror_position[1], mirror_position[2], mirror_normal, .005)

    origins = torch.randn(2, 3)
    directions = [(mirror_position[0].item() - origins[:, 0]).reshape(-1, 1),
                  (mirror_position[1].item() - origins[:, 1]).reshape(-1, 1),
                  (mirror_position[2].item() - origins[:, 2]).reshape(-1, 1)]
    directions = torch.cat(directions, dim=1)
    rays = optics.Rays(origins, directions)
    t = mirror.get_ray_intersection(rays)
    outgoing_rays, _ = mirror.intersect(rays, t)

    # Toy loss
    loss = (outgoing_rays.origins * outgoing_rays.directions).sum()
    loss.backward()

    assert mirror_normal.is_leaf
    assert mirror_position.is_leaf
    assert mirror_position.grad.abs().sum() > 0
    assert mirror_normal.grad.abs().sum() > 0
    return 0


def _test_grad_lens_wrt_self_parameters():
    return 0


def _test_grad_sensor_wrt_self_parameters():
    return 0


def _test_forward_ray_tracing(f=0.05, m=0.15, device='cpu'):
    # Creating a scene
    image_pof = -f * (1 + m)
    object_pof = f * (1 + m) / m
    lens = optics.PerfectLens(f=f, na=1 / 1.4, position=[0., 0., 0.], m=m)
    sensor = optics.Sensor(resolution=(9600, 9600), pixel_size=(3.76e-6, 3.76e-6), position=(image_pof, 0, 0),
                           poisson_noise_mean=2, quantum_efficiency=0.8)
    atom_cloud = optics.AtomCloud(n=int(1e6), f=2, position=[object_pof, 0., 0.], phi=0.1)
    light_source = optics.LightSourceFromDistribution(atom_cloud)
    scene = optics.Scene(light_source)
    scene.add_object(lens)
    scene.add_object(sensor)

    # Using the built-in function forward_ray_tracing
    rays = light_source.sample_rays(10_000_000, device=device)
    optics.forward_ray_tracing(rays, scene, max_iterations=2)

    # Readout the sensor
    produced_image = sensor.readout(add_poisson_noise=False).data.cpu().numpy()

    c = (4800, 4800)
    w = 40
    plt.imshow(produced_image[c[0] - w: c[0] + w, c[1] - w: c[1] + w], cmap='Blues')
    plt.savefig('test_forward_ray_tracing.pdf')
    plt.close()

    return 0


def _test_backward_ray_tracing(f=0.05, m=0.15, device='cpu'):
    # Creating a scene
    image_pof = -f * (1 + m)
    object_pof = f * (1 + m) / m
    lens = optics.PerfectLens(f=f, na=1 / 1.4, position=[0., 0., 0.], m=m)
    sensor = optics.Sensor(resolution=(9600, 9600), pixel_size=(3.76e-6, 3.76e-6), position=(image_pof, 0, 0),
                           poisson_noise_mean=2, quantum_efficiency=0.8)
    atom_cloud = optics.AtomCloud(n=int(1e6), f=2, position=[object_pof, 0., 0.], phi=0.1)
    light_source_bounding_shape = optics.BoundingSphere(radii=1e-3, xc=f * (1 + m) / m, yc=0.0, zc=0.0)
    light_source = optics.LightSourceFromDistribution(atom_cloud, bounding_shape=light_source_bounding_shape)
    scene = optics.Scene(light_source)
    scene.add_object(lens)
    scene.add_object(sensor)

    # Computing incident rays (assuming pinhole camera)
    N = 40
    px_j, px_i = torch.meshgrid(torch.linspace(N, -N + 1, steps=N * 2), torch.linspace(N, -N + 1, steps=N * 2))
    px_j = px_j.reshape(-1, 1).type(torch.long)
    px_i = px_i.reshape(-1, 1).type(torch.long)
    pos_x = (px_i - 0.5) * sensor.pixel_size[0]
    pos_y = (px_j - 0.5) * sensor.pixel_size[1]
    pos_z = torch.zeros(pos_x.shape)
    origins = torch.cat((pos_x, pos_y, pos_z), dim=1)
    origins = sensor.c2w.apply_transform_(origins)
    directions = optics.batch_vector(- origins[:, 0], - origins[:, 1], - origins[:, 2])
    incident_rays = optics.Rays(origins, directions, device=device)

    # Producing an image with backward ray tracing
    integrator = optics.StratifiedSamplingIntegrator(100)
    image = optics.backward_ray_tracing(incident_rays, scene, light_source, integrator, max_iterations=2)
    image = image.reshape(2 * N, 2 * N).data.cpu().numpy()

    plt.imshow(image)
    plt.savefig('test_backward_ray_tracing.pdf')
    plt.close()

    return 0


def _test_curved_mirrors():
    # Creating a scene
    f = 0.05
    m = 0.15
    lens = optics.PerfectLens(f=f, na=1 / 1.4, position=[0., 0., 0.], m=m)
    sensor = optics.Sensor(position=(-f * (1 + m), 0, 0))
    atom_cloud = optics.AtomCloud(n=int(1e6), f=2, position=[f * (1 + m) / m / 2, 0., 0.], phi=0.1)
    mirror = optics.CurvedMirror(40, .05, .05, optics.simple_transform.SimpleTransform(
        0., 0., 0., torch.tensor([f * (1 + m) / m * 3 / 4, 0, 0])))
    light_source = optics.LightSourceFromDistribution(atom_cloud)
    scene = optics.Scene(light_source)
    scene.add_object(lens)
    scene.add_object(sensor)
    scene.add_object(mirror)

    # Producing an image
    device = 'cpu'
    rays = light_source.sample_rays(10_00_000, device=device)
    rays.directions[:, 0] = rays.directions[:, 0].abs()
    optics.forward_ray_tracing(rays, scene, max_iterations=3)

    # Readout the sensor
    c = (4800, 4800)
    w = 40
    produced_image = sensor.readout(add_poisson_noise=False).data.cpu().numpy()
    plt.imshow(produced_image[c[0] - w: c[0] + w, c[1] - w: c[1] + w], cmap='Blues')
    plt.savefig('test_curved_mirrors.pdf')
    plt.close()
    return 0


"""
            Tests for pytest
"""


def test_ray():
    assert _test_ray(1) == 0
    assert _test_ray(1000) == 0


def test_rejection_sampling():
    assert _test_rejection_sampling() == 0


def test_atom_cloud():
    assert _test_atom_cloud() == 0


def test_lens():
    for lens_position in [[0., 0., 0.], [-0.1, 0., 0.], [0.1, 0.1, 0.1]]:
        assert _test_lens(f=0.05, m=0.15, position=lens_position) == 0
        assert _test_lens(f=0.04, m=0.15, position=lens_position) == 0
        assert _test_lens(f=0.01, m=0.15, position=lens_position) == 0
        assert _test_lens(f=0.05, m=0.1, position=lens_position) == 0
        assert _test_lens(f=0.04, m=0.1, position=lens_position) == 0
        assert _test_lens(f=0.01, m=0.1, position=lens_position) == 0

        assert _test_lens(f=0.05, m=0.15, right_of_lens=False, position=lens_position) == 0
        assert _test_lens(f=0.01, m=0.15, right_of_lens=False, position=lens_position) == 0
        assert _test_lens(f=0.05, m=0.1, right_of_lens=False, position=lens_position) == 0

    assert _test_lens_transform(f=0.05, m=0.15, position=[0., 0., 0.]) == 0


def test_thick_lens():
    assert _test_thick_lens() == 0


def test_ray_marching():
    assert _ray_marching_test_sensor() == 0
    assert _ray_marching_test_lens() == 0


def test_window():
    assert _test_window() == 0


def test_gradients_wrt_incident_rays():
    """
    For each optical component, check that the gradients with respect the incident rays propagate correctly
    """
    assert _test_grad_rays() == 0
    assert _test_grad_mirror_wrt_incident_rays() == 0
    assert _test_grad_lens_wrt_incident_rays() == 0
    # assert _test_grad_sensor_wrt_incident_rays() == 0


def test_gradients_wrt_self_parameters():
    """
    For each optical component, check that the gradients with respect its parameters propagate correctly
    """
    assert _test_grad_mirror_wrt_self_parameters() == 0
    assert _test_grad_lens_wrt_self_parameters() == 0
    assert _test_grad_sensor_wrt_self_parameters() == 0


def test_vectors():
    assert _test_cross_product() == 0


def test_transforms():
    assert _test_look_at_transform() == 0


def test_bounding_sphere():
    assert _test_bounding_sphere() == 0


def test_forward_ray_tracing():
    assert _test_forward_ray_tracing() == 0


def test_backward_ray_tracing():
    assert _test_backward_ray_tracing() == 0


def test_curved_mirrors():
    assert _test_curved_mirrors() == 0



import abc  # Abstract Base Classes


class BaseLightSource(abc.ABC):
    """
    Base class for light sources.
    """

    def __init__(self, bounding_shape=None):
        """
        :param bounding_shape: A bounding shape that bounds the light source
                               (:py:class:`~gradoptics.optics.bounding_shape.BoundingShape`). Default is ``None``

        .. note::
             A bounding shape is required if this light source is used with backward ray tracing

        """
        self.bounding_shape = bounding_shape

    @abc.abstractmethod
    def sample_rays(self, nb_rays, device='cpu'):
        """
        Samples rays from the light source

        :param nb_rays: Number of rays to sample (:obj:`int`)
        :param device: The desired device of returned tensor (:obj:`str`). Default is ``'cpu'``

        :return: Sampled rays (:py:class:`~gradoptics.optics.ray.Rays`)
        """
        return NotImplemented

    @abc.abstractmethod
    def plot(self, ax):
        """
        Plots the position of the light source on the provided axes.

        :param ax: 3d axes (:py:class:`mpl_toolkits.mplot3d.axes3d.Axes3D`)
        """
        return NotImplemented



import torch
from gradoptics.integrator.base_integrator import BaseIntegrator


class StratifiedSamplingIntegrator(BaseIntegrator):
    """
    Computes line integrals using stratified sampling
    """

    def __init__(self, nb_mc_steps):
        """
        :param nb_mc_steps: Number of Monte Carlo integration steps used for approximating the integral (:obj:`int`)
        """
        self.nb_mc_steps = nb_mc_steps

    def compute_integral(self, incident_rays, pdf, t_min, t_max):

        # Interpolate t between t_min and t_max
        t = torch.linspace(0, 1, self.nb_mc_steps + 1, device=incident_rays.origins.device)
        t = (t_max - t_min).expand((self.nb_mc_steps + 1, t_max.shape[0])).T * t + \
            t_min.expand((self.nb_mc_steps + 1, t_max.shape[0])).T  # [nb_rays, nb_MC_steps]

        # Perturb sampling along each ray
        mid = (t[:, :-1] + t[:, 1:]) / 2.
        lower = torch.cat((t[:, :1], mid), -1)
        upper = torch.cat((mid, t[:, -1:]), -1)
        u = torch.rand(t.shape, device=incident_rays.origins.device)
        t = lower + (upper - lower) * u  # [nb_rays, nb_MC_steps]
        delta = t[:, 1:] - t[:, :-1]

        # 3d positions at the different times t
        x = incident_rays.origins.expand(self.nb_mc_steps, -1, -1).transpose(0, 1) + mid.unsqueeze(
            -1) * incident_rays.directions.expand(self.nb_mc_steps, -1, -1).transpose(0, 1)

        densities = pdf(x.reshape(-1, 3)).reshape((x.shape[:2]))

        return (densities * delta).sum(dim=1)



import numpy as np
import torch

from gradoptics.optics.ray import Rays
from gradoptics.optics.bounding_shape import BoundingShape


class BoundingSphere(BoundingShape):
    """
    Models a sphere that can be used to bound a volume or a scene. Intersections with the sphere will not modify the
    orientation of the incoming rays (no reflection nor refraction).
    """

    def __init__(self, radii=1e-3, xc=0.31, yc=0.0, zc=0.0):
        """
        :param radii: Radii of the sphere (:obj:`float`)
        :param xc: Position of the center of the sphere along the x axis (:obj:`float`)
        :param yc: Position of the center of the sphere along the y axis (:obj:`float`)
        :param zc: Position of the center of the sphere along the z axis (:obj:`float`)
        """

        super().__init__()
        self.radii = radii
        self.xc = xc
        self.yc = yc
        self.zc = zc

    def get_ray_intersection_(self, incident_rays, eps=1e-15):

        # Computes the intersection of the incident_ray with the sphere
        origins = incident_rays.origins
        directions = incident_rays.directions
        # Solve the quadratic equation a . t^2 + b . t + c = 0 for t
        # <=> (o_x + t * d_x - self.xc)**2 + (o_y + t * d_y - self.yc)**2 +
        #     (o_z + t * d_z - self.zc)**2 = r**2
        a = directions[:, 0] ** 2 + directions[:, 1] ** 2 + directions[:, 2] ** 2
        b = 2 * ((directions[:, 0] * (origins[:, 0] - self.xc))
                 + (directions[:, 1] * (origins[:, 1] - self.yc))
                 + (directions[:, 2] * (origins[:, 2] - self.zc)))
        c = (origins[:, 0] - self.xc) ** 2 + (origins[:, 1] - self.yc) ** 2 + (
                origins[:, 2] - self.zc) ** 2 - self.radii ** 2
        pho = b ** 2 - 4 * a * c

        # Initialize return values
        t_min = torch.zeros(origins.shape[0], device=incident_rays.device, dtype=origins.dtype) + float('nan')
        t_max = torch.zeros(origins.shape[0], device=incident_rays.device, dtype=origins.dtype) + float('nan')

        # Keep the rays that intersect with the sphere
        cond = pho > 0
        t_1 = (- b[cond] + torch.sqrt(pho[cond])) / (2 * a[cond])
        t_2 = (- b[cond] - torch.sqrt(pho[cond])) / (2 * a[cond])

        t_min[cond] = torch.min(t_1, t_2)
        t_max[cond] = torch.max(t_1, t_2)

        # Check that t is greater than 0
        cond_gt_0 = cond & (t_min > eps)
        t_min[~cond_gt_0] = t_max[~cond_gt_0]  # Update negative t with t_max (potentially > 0)
        # Remove self-intersections
        cond_gt_0 = cond & (t_min > eps)
        t_min[~cond_gt_0] = float('nan')
        t_max[~cond_gt_0] = float('nan')

        return t_min, t_max

    def get_ray_intersection(self, incident_rays, eps=1e-15):
        return self.get_ray_intersection_(incident_rays, eps=eps)[0]

    def intersect(self, incident_rays, t):

        origins = incident_rays.origins
        directions = incident_rays.directions

        # Update the origin of the incoming rays
        origins = origins + t.unsqueeze(1) * directions

        return (Rays(origins, directions, luminosities=incident_rays.luminosities, device=incident_rays.device,
                meta=incident_rays.meta), torch.ones(origins.shape[0], dtype=torch.bool, device=origins.device))

    def plot(self, ax, color='grey', alpha=0.4):
        u = np.linspace(0, 2 * np.pi, 100)
        v = np.linspace(0, np.pi, 100)
        x = self.radii * np.outer(np.cos(u), np.sin(v)) + self.xc
        y = self.radii * np.outer(np.sin(u), np.sin(v)) + self.yc
        z = self.radii * np.outer(np.ones(np.size(u)), np.cos(v)) + self.zc
        ax.plot_surface(x, y, z, color=color, alpha=alpha)



from gradoptics.optics.base_optics import BaseOptics
from gradoptics.optics.ray import Rays
import torch


class Camera(BaseOptics):
    """
    Models a camera by compounding multiple optical elements.
    """

    def __init__(self, lens, sensor, intermediate_objects=[]):
        """
        :param lens: External lens - interface with the world (:py:class:`~gradoptics.optics.lens.Lens`)
        :param sensor: Sensor (:py:class:`~gradoptics.optics.sensor.Sensor`)
        :param intermediate_objects: list of optical elements between the external lens and the sensor
                                     (:obj:`list` of :py:class:`~gradoptics.optics.base_optics.BaseOptics`)
        """
        super(Camera, self).__init__()
        self.lens = lens
        self.sensor = sensor
        self.intermediate_objects = intermediate_objects

    def get_ray_intersection(self, incident_rays):
        return self.lens.get_ray_intersection(incident_rays)

    def intersect(self, incident_rays, t):
        rays = self.lens.intersect(incident_rays, t)
        for obj in self.intermediate_objects:
            t = obj.get_ray_intersection(rays)
            rays = obj.intersect(rays, t)
        t = self.sensor.get_ray_intersection(rays)
        self.sensor.intersect(rays, t, do_pixelize=True, quantum_efficiency=True)

        # Returns nan rays
        origins = torch.zeros_like(incident_rays.origins) + float('nan')
        directions = torch.zeros_like(incident_rays.directions) + float('nan')
        luminosities = (torch.zeros_like(incident_rays.luminosities) + float('nan'))
        meta = {}
        for key in incident_rays.meta.keys():
            meta[key] = torch.zeros_like(incident_rays.meta[key]) + float('nan')

        mask = torch.zeros(origins.shape[0], dtype=torch.bool, device=origins.device)  # No rays reflected or refracted
        return Rays(origins, directions, luminosities=luminosities, device=incident_rays.device, meta=meta), mask

    def plot(self, ax):
        self.lens.plot(ax)
        for obj in self.intermediate_objects:
            obj.plot(ax)
        self.sensor.plot(ax)



import torch

from gradoptics.transforms.base_transform import BaseTransform


class SimpleTransform(BaseTransform):
    """
    Transform characterized by rotation vector [w_x, w_y, w_z]
    """

    def __init__(self, theta_x, theta_y, theta_z, t):
        """
        :param theta_x: Rotation along the x axis (:obj:`float`)
        :param theta_y: Rotation along the y axis (:obj:`float`)
        :param theta_z: Rotation along the z axis (:obj:`float`)
        :param t: Translation vector, i.e. position of the transform (:obj:`torch.tensor`)
        """

        M_x = torch.tensor([[1, 0, 0],
                            [0, torch.cos(torch.tensor([theta_x])), -torch.sin(torch.tensor([theta_x]))],
                            [0, torch.sin(torch.tensor([theta_x])), torch.cos(torch.tensor([theta_x]))]],
                           dtype=torch.double)

        M_y = torch.tensor([[torch.cos(torch.tensor([theta_y])), 0, -torch.sin(torch.tensor([theta_y]))],
                            [0, 1, 0],
                            [torch.sin(torch.tensor([theta_y])), 0, torch.cos(torch.tensor([theta_y]))]],
                           dtype=torch.double)

        M_z = torch.tensor([[torch.cos(torch.tensor([theta_z])), -torch.sin(torch.tensor([theta_z])), 0],
                            [torch.sin(torch.tensor([theta_z])), torch.cos(torch.tensor([theta_z])), 0],
                            [0, 0, 1]],
                           dtype=torch.double)
        M = M_z @ M_y @ M_x

        self.transform = torch.tensor([[M[0, 0], M[0, 1], M[0, 2], t[0]],
                                       [M[1, 0], M[1, 1], M[1, 2], t[1]],
                                       [M[2, 0], M[2, 1], M[2, 2], t[2]],
                                       [0, 0, 0, 1]], dtype=torch.double)
        self.inverse_transform = torch.inverse(self.transform)

        # Sanity check (making sure that the inversion went well)
        assert torch.allclose(torch.matmul(self.transform, self.inverse_transform), torch.eye(4, dtype=torch.double),
                              rtol=1e-04, atol=1e-06)
        x = torch.randn((4, 4), dtype=torch.double)
        assert torch.allclose(torch.matmul(self.transform, torch.matmul(self.inverse_transform, x)), x, rtol=1e-04,
                              atol=1e-06)
        assert torch.allclose(torch.matmul(self.transform, torch.matmul(self.inverse_transform, x)), x, rtol=1e-04,
                              atol=1e-06)



import torch


class PSF(torch.nn.Module):
    """
    Models a point spread function.
    """

    def __init__(self, kernel):
        """
        :param kernel: Kernel that characterizes the point spread function (:obj:`torch.tensor`)
        """
        super(PSF, self).__init__()
        self.kernel = kernel

    def forward(self, x):
        """
        Performs a convolution between the input data and the kernel of the point spread function

        :param x: Data to convolve with the PSF (:obj:`torch.tensor`)

        :return: The convolved input (:obj:`torch.tensor`)
        """
        # Fourier convolution
        fft_x = torch.fft.fftn(x)
        fft_kernel = torch.fft.fftn(self.kernel.to(x.device), s=fft_x.shape)

        # Multiplication and inverse fourier transform
        output = torch.fft.ifftn(fft_x * fft_kernel)

        # Discard complex parts
        return output.real



import torch
import warnings
import gradoptics as optics


def trace_rays(incident_rays, scene):
    """
    Performs one iteration of ray tracing, i.e. finds the closest object in the ``scene`` each ray will intersect, and
    computes the reflected or refracted rays. Rays with ``nan`` components are returned for rays that do not have
    intersections

    :param incident_rays: Batch of incident rays (:py:class:`~gradoptics.optics.ray.Rays`)
    :param scene: Scene in which the rays are travelling (:py:class:`~gradoptics.ray_tracing.scene.Scene`)

    :return: (:obj:`tuple`)
             - Reflected or refracted rays after one iteration of ray tracing (:py:class:`~gradoptics.optics.ray.Rays`)
             - Times at which the rays intersect the closest object to them and ``nan`` for rays that have no
               intersection (:obj:`torch.tensor`)
             - Boolean tensor that indicates if the ``incident_rays`` have been reflected/refracted or no
               (:obj:`torch.tensor`)
    """
    device = incident_rays.device
    t = torch.zeros(incident_rays.origins.shape[0], device=device) + float('Inf')
    outgoing_rays = optics.empty_like(incident_rays)
    obj_indices = torch.zeros(incident_rays.origins.shape[0], device=device, dtype=torch.long) - 1
    mask = torch.zeros(incident_rays.origins.shape[0], device=device, dtype=torch.bool)

    # Find first object intersected by each ray
    for idx, o in enumerate(scene.objects):
        t_current = o.get_ray_intersection(incident_rays)
        # If there is an intersection and if the intersected object is the first one the ray intersects, update the
        # return values
        condition = (t_current < t) & (t_current > 1e-3) & (~torch.isnan(t_current))
        t[condition] = t_current[condition].type(t.dtype)
        obj_indices[condition] = idx
        mask[condition] = 1

    # Compute the rays reflected / refracted by their intersections with the objects in the scene
    for idx, o in enumerate(scene.objects):
        condition = (obj_indices == idx)

        if condition.sum() == 0:
            continue

        tmp, mask[condition] = o.intersect(incident_rays[condition], t[condition])

        if tmp is not None:
            outgoing_rays[condition & mask] = tmp

    return outgoing_rays, t, mask


def forward_ray_tracing(incident_rays, scene, max_iterations=2, ax=None):
    """
    Performs forward ray tracing, i.e. computes the path taken by the rays ``incident_rays`` in the scene ``scene``
    until the maximum number of bounces ``max_iterations`` is reached, or until the rays hit a sensor

    :param incident_rays: Rays that should be traced in the scene (:py:class:`~gradoptics.optics.ray.Rays`)
    :param scene: Scene in which the rays are travelling (:py:class:`~gradoptics.ray_tracing.scene.Scene`)
    :param max_iterations: Maximum number of bounces in the ``scene`` (:obj:`int`)
    :param ax: 3d axes in which the rays are plotted (if ``ax`` is not ``None``) as they traverse the scene
               (:py:class:`mpl_toolkits.mplot3d.axes3d.Axes3D`). Default is ``None``
    """

    for i in range(max_iterations):
        outgoing_rays, t, mask = trace_rays(incident_rays, scene)

        if ax is not None:
            origins = incident_rays.origins[mask].data.cpu().numpy()
            destinations = incident_rays[mask](t[mask]).data.cpu().numpy()
            for ray_idx in range(origins.shape[0]):
                ax.plot([origins[ray_idx, 0], destinations[ray_idx, 0]],
                        [origins[ray_idx, 1], destinations[ray_idx, 1]],
                        [origins[ray_idx, 2], destinations[ray_idx, 2]])

        # Only keep the rays that intersect with the system
        incident_rays = outgoing_rays[mask]

        torch.cuda.empty_cache()

    if incident_rays.origins.shape[0] > 0:
        warnings.warn("The maximum number of iteration has been reached and there are still rays"
                      "bouncing in the scene. incident_rays.shape[0]: " + str(incident_rays.origins.shape[0]))


def backward_ray_tracing(incident_rays, scene, light_source, integrator, max_iterations=2):
    """
    Performs backward ray tracing, i.e. computes the path taken by the rays ``incident_rays`` in the scene ``scene``
    until the maximum number of bounces ``max_iterations`` is reached, or until the rays hit the light source
    ``light_source``

    :param incident_rays: Rays that should be traced in the scene (:py:class:`~gradoptics.optics.ray.Rays`)
    :param scene: Scene in which the rays are travelling (:py:class:`~gradoptics.ray_tracing.scene.Scene`)
    :param light_source: Light source (:py:class:`~gradoptics.light_sources.base_light_source.BaseLightSource`)
    :param integrator: An integrator to compute line integrals
                       (:py:class:`~gradoptics.integrator.base_integrator.BaseIntegrator`)
    :param max_iterations: Maximum number of bounces in the ``scene`` (:obj:`int`)

    :return: the intensity carried by the rays ``incident_rays`` (:py:class:`~gradoptics.optics.ray.Rays`)
    """

    intensities = torch.zeros(incident_rays.get_size(), dtype=torch.double)

    # Labelling the rays
    incident_rays.meta['track_idx'] = torch.linspace(0, incident_rays.get_size() - 1, incident_rays.get_size(),
                                                     dtype=torch.long)

    for i in range(max_iterations):
        outgoing_rays, t, mask = trace_rays(incident_rays, scene)

        # Potential intersection with the light source
        t_min, t_max = light_source.bounding_shape.get_ray_intersection_(incident_rays)

        # Mask for the rays that hit the light source rather than the object found in trace_rays
        new_mask = (t_min < t) & (t_min < t_max)

        # Computing the intensities for the rays that have hit the light source
        if new_mask.sum() > 0:
            intensities[incident_rays.meta['track_idx'][new_mask]] = integrator.compute_integral(
                incident_rays[new_mask], light_source.pdf, t_min[new_mask], t_max[new_mask])

        # Rays that are still in the scene, and have not hit the light source
        incident_rays = outgoing_rays[mask & (~new_mask)]

    return intensities



import torch
import warnings

from gradoptics.optics.base_optics import BaseOptics
from gradoptics.transforms.look_at_transform import LookAtTransform


class Sensor(BaseOptics):
    """
    Models a sensor.
    """

    def __init__(self, resolution=(9600, 9600), pixel_size=(3.76e-6, 3.76e-6), position=(-0.0575, 0, 0),
                 poisson_noise_mean=2, quantum_efficiency=0.8, viewing_direction=[1., 0., 0.], up=[0., 0., 1.],
                 psfs={}, psf_ratio=1):
        """
        :param resolution: Image processing convention: origin in the upper left corner, horizontal x axis and vertical
                           y axis (:obj:`tuple`)
        :param pixel_size: Image processing convention: origin in the upper left corner, horizontal x axis and vertical
                           y axis  (:obj:`tuple`)
        :param position: Position of the lens  (:obj:`list`)
        :param poisson_noise_mean: Mean readout noise  (:obj:`float`)
        :param quantum_efficiency: Quantum efficiency (:obj:`float`)
        :param viewing_direction: Viewing direction of the sensor (:obj:`list`)
        :param up: A vector that orients the sensor with respect to the viewing direction (:obj:`list`)
                   For example, if up=[0., 0., 1.] the top of the camera will point upwards.
                   If up=[0., 0., -1.], the top of the camera will point downwards.
        :param psfs: Depth and position-dependant point spread function (:obj:`dict`)
        :param psf_ratio: Ratio between the psf resolution and the sensor resolution (:obj:`int`)
        """
        super(Sensor, self).__init__()
        self.position = position
        self.resolution = resolution
        self.pixel_size = pixel_size
        self.poisson_noise_mean = poisson_noise_mean
        self.quantum_efficiency = quantum_efficiency
        self.c2w = LookAtTransform(torch.tensor(viewing_direction), position, up=torch.tensor(up))
        self.viewing_direction = torch.tensor(viewing_direction)
        self.position = position
        self.add_psf = len(psfs.keys()) > 0
        self.psfs = psfs
        self.psf_ratio = psf_ratio
        self.resolution = resolution
        assert type(psf_ratio) == int

        if self.add_psf:
            self.depth_images = [torch.zeros((resolution[1] * psf_ratio, resolution[0] * psf_ratio, 1)) for _ in
                                 range(len(psfs['data'].keys()))]
            self.psf_depth_bounds = list(psfs['data'].keys())
        else:
            self.depth_images = [torch.zeros((resolution[1] * psf_ratio, resolution[0] * psf_ratio, 1))]

    def get_ray_intersection(self, incident_rays):

        # World space to camera space
        incident_rays = self.c2w.apply_inverse_transform(incident_rays)

        origins = incident_rays.origins
        directions = incident_rays.directions
        # The optical axis is the z axis in the camera space
        # Find the intersection with the sensor plane
        t = - origins[:, 2] / directions[:, 2]
        x = origins[:, 0] + t * directions[:, 0]
        y = origins[:, 1] + t * directions[:, 1]

        eps_x, eps_y = self.pixel_size[0] / 5, self.pixel_size[1] / 5
        condition = (t > 0) & ((torch.abs(x) + eps_x) < self.resolution[0] / 2 * self.pixel_size[0]) & \
                    ((torch.abs(y) + eps_y) < self.resolution[1] / 2 * self.pixel_size[1])
        t[~condition] = float('nan')
        return t

    def intersect(self, incident_rays, t, do_pixelize=True, quantum_efficiency=True):

        # World space to camera space
        incident_rays = self.c2w.apply_inverse_transform(incident_rays)

        origins = incident_rays.origins
        directions = incident_rays.directions
        hit_positions = origins + t.unsqueeze(1) * directions
        assert torch.allclose(hit_positions[:, 2], torch.zeros(1, device=hit_positions.device,
                                                               dtype=hit_positions.dtype), atol=1e-06)

        # Camera space (origin in the center of the image, horizontal x axis pointing to the right and vertical y axis
        # pointing upwards) to python convention (origin in the upper left, vertical x axis and horizontal y axis)
        nb_horizontal_pixel_from_center = hit_positions[:, 0] / (self.pixel_size[1])
        nb_vertical_pixel_from_center = hit_positions[:, 1] / (self.pixel_size[0])
        hit_positions[:, 0] = -nb_vertical_pixel_from_center + self.resolution[1] // 2
        hit_positions[:, 1] = -nb_horizontal_pixel_from_center + self.resolution[0] // 2

        if do_pixelize:

            if self.add_psf:

                nb_processed_rays = 0
                for depth_id, (low_bounds, high_bounds) in enumerate(self.psf_depth_bounds):
                    # Get the mask for the rays at the depths in [low_bounds, high_bounds]
                    mask = (incident_rays.meta['depth'] < high_bounds) & (incident_rays.meta['depth'] > low_bounds)
                    hit_positions_psf = hit_positions[mask] * self.psf_ratio
                    self.pixelize(depth_id, hit_positions_psf, incident_rays.luminosities[mask],
                                  quantum_efficiency=quantum_efficiency)
                    nb_processed_rays += mask.sum()

                if not (nb_processed_rays == hit_positions.shape[0]):
                    raise Exception(
                        "Some rays were not processed: their origins' depths were not included in the psfs.")

            else:
                self.pixelize(0, hit_positions, incident_rays.luminosities, quantum_efficiency=quantum_efficiency)

        # No rays reflected or refracted
        return None, torch.zeros(origins.shape[0], dtype=torch.bool, device=origins.device)

    def pixelize(self, depth_id, hit_positions, luminosities, quantum_efficiency=True):
        """
        Accumulates photons at the pixels hit by some rays at the positions ``hit_positions``

        :param depth_id: index of the depth (encoded in the PSF) from where the rays come from (:obj:`int`)
        :param hit_positions: The positions at which the rays hit the sensor (:obj:`torch.tensor`)
        :param luminosities: The luminosities of the rays  (:obj:`torch.tensor`)
        :param quantum_efficiency: Whether to use quantum efficiency or no (:obj:`bool`)
        """

        self.depth_images[depth_id] = self.depth_images[depth_id].to(hit_positions.device)

        if hit_positions.shape[0] == 0:
            return

        if quantum_efficiency:
            luminosities *= self.quantum_efficiency

        # Update pixel values
        indices = torch.floor(hit_positions[:, 1].type(torch.float64)).type(torch.int64) * (
                    self.resolution[1] * self.psf_ratio) + \
                  torch.floor(hit_positions[:, 0].type(torch.float64)).type(torch.int64)
        max_nb_identical_indices = indices.unique(return_counts=True)[1].max()
        sorted_indices, arg_sorted_indices = torch.sort(indices)

        for i in range(max_nb_identical_indices):
            ind = sorted_indices[i::max_nb_identical_indices]
            arg_ind = arg_sorted_indices[i::max_nb_identical_indices]

            # Sanity check, should be removed
            cond = (ind[1:] == ind[:-1])
            assert cond.sum().item() == 0, "Should not happen"

            self.depth_images[depth_id][ind % (self.resolution[1] * self.psf_ratio),
                                        ind // (self.resolution[1] * self.psf_ratio)] += luminosities[arg_ind, None]

    def readout(self, add_poisson_noise=True, destructive_readout=True):
        """
        Readouts the sensor

        :param add_poisson_noise: Whether to add poisson noise or no (:obj:`bool`)
        :param destructive_readout: Whether the accumulated photons should be destroyed or no (:obj:`bool`)

        :returns: The produced image (:obj:`torch.tensor`)
        """
        # If psfs were specified
        if self.add_psf:
            # Apply psf to each depth_image
            for i, depth in enumerate(self.psfs['data'].keys()):
                for key in self.psfs['data'][depth].keys():
                    self.depth_images[i][key[0]:key[1], :, 0] = self.psfs['data'][depth][key](
                        self.depth_images[i][key[0]:key[1], :, 0])

        # Summing all the depth images & rebinning
        image = (torch.nn.AvgPool2d(self.psf_ratio)(
            torch.cat(self.depth_images, dim=2).sum(dim=2).unsqueeze(0).unsqueeze(0)) * self.psf_ratio ** 2).squeeze(
            0).squeeze(0)

        if destructive_readout:
            # Reinitialize depth images
            for i in range(len(self.depth_images)):
                self.depth_images[i] = self.depth_images[i].detach() * 0

        # Add readout noise
        if add_poisson_noise:
            image = image + torch.poisson(torch.zeros_like(image) + self.poisson_noise_mean, generator=None)

        return image

    def plot(self, ax, s=0.1, color='grey', resolution_=100):

        # @Todo, change this to plot_surface
        res = self.resolution[0] * self.pixel_size[0] / resolution_
        x = torch.arange(-self.resolution[0] * self.pixel_size[0] / 2,
                         self.resolution[0] * self.pixel_size[0] / 2,
                         res)
        y = torch.arange(-self.resolution[1] * self.pixel_size[1] / 2,
                         self.resolution[1] * self.pixel_size[1] / 2,
                         res)

        x, y = torch.meshgrid(x, y)

        x = x.reshape(-1)
        y = y.reshape(-1)
        z = torch.zeros(x.shape)

        # Lens space to world space
        xyz = self.c2w.apply_transform_(torch.cat((x.reshape(-1, 1), y.reshape(-1, 1), z.reshape(-1, 1)), dim=1))
        ax.scatter(xyz[:, 0], xyz[:, 1], xyz[:, 2], s=s, c=color)

    # def plot_image(self, ax, img_height=750, cmap="Blues"):
    #    raise NotImplementedError("Not implemented yet.")

    def sample_points_on_sensor(self, nb_points, device='cpu'):
        """
        Sample points uniformly on the sensor

        :param nb_points: Number of points to sample (:obj:`int`)
        :param device: The desired device of returned tensor (:obj:`str`). Default is ``'cpu'``

        :return: Sampled points (:obj:`torch.tensor`)
        """

        points = torch.zeros((nb_points, 3), device=device)
        points[:, 0] = torch.rand(nb_points, device=device) * (self.pixel_size[0] * self.resolution[0]) - (
                self.pixel_size[0] * self.resolution[0] / 2)
        points[:, 1] = torch.rand(nb_points, device=device) * (self.pixel_size[1] * self.resolution[1]) - (
                self.pixel_size[1] * self.resolution[1] / 2)
        return self.c2w.apply_transform_(points)



import abc  # Abstract Base Classes
import torch
from gradoptics.optics.ray import Rays


class BaseTransform(abc.ABC):
    """
    Base class for transforms. Enables to switch between world-space and object-space.
    """

    def apply_transform(self, rays):
        """
        Transform the coordinates of ``rays`` in world-space to coordinates in object-space

        :param rays: Rays in world-space (:py:class:`~gradoptics.optics.Ray.Rays`)

        :return: New rays in object-space (:py:class:`~gradoptics.optics.Ray.Rays`)
        """
        new_o = torch.matmul(self.transform.to(rays.device),
                             torch.cat((rays.origins.type(torch.double), torch.ones((rays.origins.shape[0], 1),
                                                                                    device=rays.device)),
                                       dim=1).unsqueeze(-1))[:, :3, 0].type(rays.origins.dtype)
        new_d = torch.matmul(self.transform.to(rays.device),
                             torch.cat((rays.directions.type(torch.double),
                                        torch.zeros((rays.directions.shape[0], 1), dtype=torch.double,
                                                    device=rays.device)), dim=1).unsqueeze(-1))[:, :3, 0].type(
            rays.directions.dtype)
        return Rays(new_o, new_d, luminosities=rays.luminosities, meta=rays.meta, device=rays.device)

    def apply_inverse_transform(self, rays):
        """
        Transform the coordinates of ``rays`` in object-space to coordinates in world-space

        :param rays: Rays in object-space (:py:class:`~gradoptics.optics.Ray.Rays`)

        :return: New rays in world-space (:py:class:`~gradoptics.optics.Ray.Rays`)
        """
        new_o = torch.matmul(self.inverse_transform.to(rays.device),
                             torch.cat((rays.origins.type(torch.double),
                                        torch.ones((rays.origins.shape[0], 1), dtype=torch.double,
                                                   device=rays.device)), dim=1).unsqueeze(-1))[:, :3, 0].type(
            rays.origins.dtype)
        new_d = torch.matmul(self.inverse_transform.to(rays.device),
                             torch.cat((rays.directions.type(torch.double), torch.zeros((rays.directions.shape[0], 1),
                                                                                        device=rays.device)),
                                       dim=1).unsqueeze(-1))[:, :3, 0].type(rays.directions.dtype)
        return Rays(new_o, new_d, luminosities=rays.luminosities, meta=rays.meta, device=rays.device)

    def apply_transform_(self, points):
        """
        Transform the coordinates of ``points`` in world-space to coordinates in object-space

        :param points: Vectors in world-space (:obj:`torch.tensor`)

        :return: New vectors in object-space (:obj:`torch.tensor`)
        """
        return torch.matmul(self.transform.to(points.device),
                            torch.cat((points.type(torch.double), torch.ones((points.shape[0], 1), dtype=torch.double,
                                                                             device=points.device)), dim=1).unsqueeze(
                                -1))[:, :3, 0].type(points.dtype)

    def apply_inverse_transform_(self, points):
        """
        Transform the coordinates of ``points`` in object-space to coordinates in world-space

        :param points: Vectors in object-space (:obj:`torch.tensor`)

        :return: New vectors in world-space (:obj:`torch.tensor`)
        """
        return torch.matmul(self.inverse_transform.to(points.device),
                            torch.cat((points.type(torch.double), torch.ones((points.shape[0], 1), dtype=torch.double,
                                                                             device=points.device)), dim=1).unsqueeze(
                                -1))[:, :3, 0].type(points.dtype)



import torch
import math
import numpy as np

from gradoptics.distributions.base_distribution import BaseDistribution


class GaussianDistribution(BaseDistribution):
    """
    1D Gaussian Distribution.
    """

    def __init__(self, mean=0.0, std=1.0):
        """
        :param mean: mean of the distribution (:obj:`float`)
        :param std: standard deviation of the distribution (:obj:`float`)
        """
        super().__init__()
        self.mean = mean
        self.std = std

    def sample(self, nb_points, device='cpu'):
        return self.mean + self.std * torch.randn(nb_points, device=device)

    def pdf(self, x):
        """
        Returns the pdf function evaluated at ``x``

        :param x: Value where the pdf should be evaluated (:obj:`torch.tensor`)

        :return: The pdf function evaluated at ``x`` (:obj:`torch.tensor`)
        """
        return 1 / (self.std * np.sqrt(2 * math.pi)) * torch.exp((- 1 / 2) * ((x - self.mean) / self.std) ** 2)



import setuptools
from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()


setup(
  name='gradoptics',
  version='0.0.1',
  description='End-to-end differentiable optics',
  long_description=long_description,
  long_description_content_type="text/markdown",
  author='Michael Kagan, Maxime Vandegar',
  author_email='makagan@slac.stanford.edu, maxime.vandegar@slac.stanford.edu',
  url="https://github.com/magis-slac/gradoptics",
  classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
  package_dir={'': 'src'},
  packages=setuptools.find_packages(where="src"),
)



import math
import abc

import torch
import numpy as np
import gradoptics as optics
from gradoptics.optics.base_optics import BaseOptics
from gradoptics.optics.ray import Rays
from gradoptics.optics.vector import batch_vector


class BaseMirror(BaseOptics):

    def __init__(self, transform=None):
        """
        :param transform: Transform to orient the mirror
                          (:py:class:`~gradoptics.transforms.base_transform.BaseTransform`)
        """
        # Identity transform
        if transform is None:
            transform = optics.simple_transform.SimpleTransform(0., 0., 0., torch.tensor([0, 0, 0]))

        self.transform = transform

    @abc.abstractmethod
    def _get_normal(self, x):
        return NotImplemented

    def intersect(self, incident_rays, t):
        """
        Returns the ray reflected by the mirror
        :param incident_rays:
        :param t:
        :return:
        """

        incident_rays = self.transform.apply_inverse_transform(incident_rays)  # World space to object space
        directions = incident_rays.directions
        collision_points = incident_rays(t)

        normal = self._get_normal(collision_points)

        # https://www.fabrizioduroni.it/2017/08/25/how-to-calculate-reflection-vector.html
        directions = -directions
        condition = torch.arccos(directions[:, 0] * normal[:, 0] + directions[:, 1] * normal[:, 1] +
                                 directions[:, 2] * normal[:, 2]) < math.pi / 2

        # It is assumed that both sides of the mirror are reflective. Inverting the normal as a function of
        # which surface was hit by the rays
        normal_sign_factor = torch.ones((directions.shape[0], 1), device=directions.device)
        normal_sign_factor[~condition.detach()] = -1
        normal = normal_sign_factor * normal

        scaling = 2 * (
                directions[:, 0] * normal[:, 0] + directions[:, 1] * normal[:, 1] + directions[:, 2] * normal[:, 2])
        direction_reflected_rays = batch_vector(scaling * normal[:, 0] - directions[:, 0],
                                                scaling * normal[:, 1] - directions[:, 1],
                                                scaling * normal[:, 2] - directions[:, 2])
        reflected_ray = self.transform.apply_transform(
            Rays(collision_points, direction_reflected_rays, luminosities=incident_rays.luminosities,
                 meta=incident_rays.meta, device=incident_rays.device))
        return reflected_ray, torch.ones(collision_points.shape[0], dtype=torch.bool, device=collision_points.device)


class FlatMirror(BaseMirror):
    """
    Models a flat mirror.
    """

    def __init__(self, x_mirror, y_mirror, z_mirror, normal, mirror_radii):
        """
        :param x_mirror: Position of the center of the mirror along the x axis (:obj:`float`)
        :param y_mirror: Position of the center of the mirror along the y axis (:obj:`float`)
        :param z_mirror: Position of the center of the mirror along the z axis (:obj:`float`)
        :param normal: Normal of the mirror (:obj:`torch.tensor`)
        :param mirror_radii: Radii of the mirror (:obj:`float`)
        """
        super().__init__()
        self.x_mirror = x_mirror
        self.y_mirror = y_mirror
        self.z_mirror = z_mirror
        self.normal = normal
        self.mirror_radii = mirror_radii

    def _get_plane_equation(self):
        """
        Returns the coefficients of the plane of the mirror ax + by + cz + d = 0
        :return: the coefficients a, b, c and d
        """

        # The equation of a plane containing the point (xm, ym, zm) with normal vector <A, B, C>
        # is given by A(x-xm) + B(y-ym) + C(z-zm) = 0
        a = self.normal[0]
        b = self.normal[1]
        c = self.normal[2]
        d = - self.normal[0] * self.x_mirror - self.normal[1] * self.y_mirror - self.normal[2] * self.z_mirror
        return a, b, c, d

    def _get_normal(self, x):
        return self.normal.expand(x.shape).to(x.device)

    def get_ray_intersection(self, incident_rays, eps=1e-15):
        # Computes the intersection of the incident_ray with the mirror plane
        origins = incident_rays.origins
        directions = incident_rays.directions
        mirror_plane_coefficients = self._get_plane_equation()
        num = -(mirror_plane_coefficients[0] * origins[:, 0] + mirror_plane_coefficients[1] * origins[:, 1] +
                mirror_plane_coefficients[2] * origins[:, 2] + mirror_plane_coefficients[3])
        den = mirror_plane_coefficients[0] * directions[:, 0] + mirror_plane_coefficients[1] * directions[:, 1] + \
            mirror_plane_coefficients[2] * directions[:, 2]
        t = num / (den + eps)

        # Make sure that the ray hits the mirror
        intersection_points = origins + t.unsqueeze(1) * directions
        condition = (t > 0) & (
                ((intersection_points[:, 0] - self.x_mirror) ** 2 + (intersection_points[:, 1] - self.y_mirror) ** 2 + (
                        intersection_points[:, 2] - self.z_mirror) ** 2) < self.mirror_radii ** 2)
        # Return nan for rays that have no intersection
        t[~condition] = float('nan')
        return t

    def plot(self, ax, show_normal=False, s=0.1, color='lightblue', resolution=100):
        y = torch.arange(-self.mirror_radii, self.mirror_radii, 2 * self.mirror_radii / resolution)
        z = torch.arange(-self.mirror_radii, self.mirror_radii, 2 * self.mirror_radii / resolution)
        y, z = torch.meshgrid(y, z)

        y = y.reshape(-1)
        z = z.reshape(-1)

        indices = (y ** 2 + z ** 2) < self.mirror_radii ** 2
        y = y[indices] + self.y_mirror
        z = z[indices] + self.z_mirror

        # Get z coordinates
        mirror_plane_coefficients = self._get_plane_equation()
        x = (-mirror_plane_coefficients[3] - mirror_plane_coefficients[2] * z - mirror_plane_coefficients[1] * y) / \
            mirror_plane_coefficients[0]

        ax.scatter(x, y, z, s=s, c=color)

        if show_normal:  # @Todo
            pass


class CurvedMirror(BaseMirror):

    def __init__(self, R, width, height, transform):
        super().__init__(transform)
        self.R = R
        self.width = width
        self.height = height

    def _get_normal(self, x):
        sphere_center = torch.tensor([-self.R, 0, 0])

        return optics.normalize_batch_vector(sphere_center - x)

    def get_ray_intersection(self, incident_rays, eps=1e-15):
        incident_rays = self.transform.apply_inverse_transform(incident_rays)  # World space to object space
        origins, directions = incident_rays.get_origin_and_direction()

        sphere_center = torch.tensor([-self.R, 0, 0])

        # Solves a quadratic equation in t (intersection between the incident_rays and the sphere)
        a = directions[:, 0] ** 2 + directions[:, 1] ** 2 + directions[:, 2] ** 2
        b = 2 * (((origins[:, 0] - sphere_center[0]) * directions[:, 0]) + (
                    (origins[:, 1] - sphere_center[1]) * directions[:, 1]) + (
                             (origins[:, 2] - sphere_center[2]) * directions[:, 2]))
        c = (origins[:, 0] - sphere_center[0]) ** 2 + (origins[:, 1] - sphere_center[1]) ** 2 + (
                    origins[:, 2] - sphere_center[2]) ** 2 - self.R ** 2

        pho = b ** 2 - 4 * a * c
        mask = pho > 0
        sqrt_pho = np.sqrt(pho[mask])

        # Roots of the quadratic equation
        t1 = torch.zeros(origins.shape[0])
        t2 = torch.zeros(origins.shape[0])
        t1[~mask] = float('Inf')
        t2[~mask] = float('Inf')
        t1[mask] = (-b[mask] + sqrt_pho) / (2 * a[mask])
        t2[mask] = (-b[mask] - sqrt_pho) / (2 * a[mask])

        # Makes sure the rays hit the sphere within the mirror width and height
        position_on_sphere = incident_rays[mask](t1[mask])
        new_mask = mask.clone()
        new_mask[mask] = (position_on_sphere[:, 1].abs() < (self.width / 2)) & (
                    position_on_sphere[:, 2].abs() < (self.height / 2))
        t1[~new_mask] = float('Inf')
        position_on_sphere = incident_rays[mask](t2[mask])
        new_mask = mask.clone()
        new_mask[mask] = (position_on_sphere[:, 1].abs() < (self.width / 2)) & (
                    position_on_sphere[:, 2].abs() < (self.height / 2))
        t2[~new_mask] = float('Inf')

        # Removes virtual intersections
        t1[t1 < 0] = float('Inf')
        t2[t2 < 0] = float('Inf')
        t = torch.min(t1, t2)
        t[torch.isinf(t)] = float('nan')

        return t

    def plot(self, ax, show_normal=False, s=0.1, color='lightblue', resolution=100):
        Y = torch.arange(-self.width / 2, self.width / 2, self.width / resolution)
        Z = torch.arange(-self.height / 2, self.height / 2, self.height / resolution)
        Y, Z = torch.meshgrid(Y, Z)
        X = torch.sqrt(self.R ** 2 - Y ** 2 - Z ** 2) - self.R

        # coordinates to world space
        xyz = self.transform.apply_transform_(torch.cat((X.reshape(-1, 1), Y.reshape(-1, 1), Z.reshape(-1, 1)), dim=1))

        # Plot the surface.
        ax.plot_surface(xyz[:, 0].reshape(X.shape).numpy(), xyz[:, 1].reshape(X.shape).numpy(),
                        xyz[:, 2].reshape(X.shape).numpy())


import torch

from gradoptics.optics.vector import normalize_vector
from gradoptics.optics.vector import cross_product
from gradoptics.transforms.base_transform import BaseTransform


class LookAtTransform(BaseTransform):
    """
    Transform that orients an object so that it is oriented towards a given direction.

    References: Physically based rendering, section 2.7.7 The look-at transform.
    """

    def __init__(self, viewing_direction, pos, up=torch.tensor([0., 0., 1.])):
        """
        Given a viewing direction in world-space, computes the 4x4 transform matrix to move a point from object-space
        to world-space and from world-space to object-space

        :param viewing_direction: viewing direction of the object (:obj:`torch.tensor`)
        :param pos: position of the object (:obj:`torch.tensor`)
        :param up: a vector that orients the object with respect to the viewing direction (:obj:`torch.tensor`).
                   For example, if up=torch.tensor([0, 0, 1]) the top of the object will point upwards.
                   If up=torch.tensor([0, 0, -1]), the top of the object will point downwards.
        """
        dir_ = normalize_vector(viewing_direction)
        left = normalize_vector(cross_product(normalize_vector(up), dir_))
        new_up = cross_product(dir_, left)

        self.transform = torch.tensor([[left[0], new_up[0], dir_[0], pos[0]],
                                       [left[1], new_up[1], dir_[1], pos[1]],
                                       [left[2], new_up[2], dir_[2], pos[2]],
                                       [0., 0., 0., 1.]], dtype=torch.double)
        self.inverse_transform = torch.inverse(self.transform)

        # Sanity check (making sure that the inversion went well)
        assert torch.allclose(torch.matmul(self.transform, self.inverse_transform), torch.eye(4, dtype=torch.double),
                              rtol=1e-04, atol=1e-06)
        x = torch.randn((4, 4), dtype=torch.double)
        assert torch.allclose(torch.matmul(self.transform, torch.matmul(self.inverse_transform, x)), x, rtol=1e-04,
                              atol=1e-06)
        assert torch.allclose(torch.matmul(self.transform, torch.matmul(self.inverse_transform, x)), x, rtol=1e-04,
                              atol=1e-06)



import torch


def batch_vector(x, y, z):
    """
    Returns a batch of vector

    :param x: Component of the vectors along the x axis (:obj:`torch.tensor`)
    :param y: Component of the vectors along the y axis (:obj:`torch.tensor`)
    :param z: Component of the vectors along the z axis (:obj:`torch.tensor`)

    :return: A batch of 3d vectors (:obj:`torch.tensor`)
    """
    return torch.cat((x.reshape(-1, 1), y.reshape(-1, 1), z.reshape(-1, 1)), dim=1)


def normalize_batch_vector(input_vector, eps=1e-15):
    """
    Normalize the ``input_vector`` using the Euclidian norm

    :param input_vector: A batch of 3d vectors (:obj:`torch.tensor`)
    :param eps: Parameter used for numerical stability. Default is ``'1e-15'``
    """
    vector_norm = torch.sqrt(input_vector[:, 0] ** 2 + input_vector[:, 1] ** 2 + input_vector[:, 2] ** 2)
    x = (input_vector[:, 0] / (vector_norm + eps)).reshape(-1, 1)
    y = (input_vector[:, 1] / (vector_norm + eps)).reshape(-1, 1)
    z = (input_vector[:, 2] / (vector_norm + eps)).reshape(-1, 1)
    return torch.cat((x, y, z), dim=1)


def dot_product(batch_vector1, batch_vector2):
    """
    Computes the dot product between the given batch of 3d vectors

    :param batch_vector1: A batch of 3d vectors (:obj:`torch.tensor`)
    :param batch_vector2: A batch of 3d vectors (:obj:`torch.tensor`)

    :return: The dot product between ``batch_vector1`` and ``batch_vector2`` (:obj:`torch.tensor`)
    """
    return batch_vector1[:, 0] * batch_vector2[:, 0] + batch_vector1[:, 1] * batch_vector2[:, 1] + batch_vector1[:, 2] \
        * batch_vector2[:, 2]


def cos_theta(batch_vector1, batch_vector2, eps=1e-15):
    """
    Computes the cosine between the given batch of 3d vectors

    :param batch_vector1: A batch of 3d vectors (:obj:`torch.tensor`)
    :param batch_vector2: A batch of 3d vectors (:obj:`torch.tensor`)
    :param eps: Parameter used for numerical stability. Default is ``'1e-15'``

    :return: The cosine between ``batch_vector1`` and ``batch_vector2`` (:obj:`torch.tensor`)
    """
    norm_v1 = torch.sqrt(batch_vector1[:, 0] ** 2 + batch_vector1[:, 1] ** 2 + batch_vector1[:, 2] ** 2)
    norm_v2 = torch.sqrt(batch_vector2[:, 0] ** 2 + batch_vector2[:, 1] ** 2 + batch_vector2[:, 2] ** 2)
    return dot_product(batch_vector1, batch_vector2) / (norm_v1 * norm_v2 + eps)


def cross_product(vector1, vector2):
    """
    Computes the cross product between the given 3d vectors

    :param vector1: A 3d vector (:obj:`torch.tensor`)
    :param vector2: A 3d vector (:obj:`torch.tensor`)

    :return: The cross product between ``vector1`` and ``vector2`` (:obj:`torch.tensor`)
    """
    x = vector1[1] * vector2[2] - vector1[2] * vector2[1]
    y = vector1[2] * vector2[0] - vector1[0] * vector2[2]
    z = vector1[0] * vector2[1] - vector1[1] * vector2[0]
    return torch.tensor([x, y, z])


def vector3d(x, y, z):
    """
    Returns a 3d vector

    :param x: Component of the vector along the x axis (:obj:`torch.tensor`)
    :param y: Component of the vector along the y axis (:obj:`torch.tensor`)
    :param z: Component of the vector along the z axis (:obj:`torch.tensor`)

    :return: A 3d vector (:obj:`torch.tensor`)
    """
    return torch.tensor([x, y, z], dtype=torch.float32)


def normalize_vector(input_vector, eps=1e-15):
    """
    Normalize the ``input_vector`` using the Euclidian norm

    :param input_vector: A 3d vector (:obj:`torch.tensor`)
    :param eps: Parameter used for numerical stability. Default is ``'1e-15'``
    """
    vector_norm = torch.sqrt(input_vector[0] ** 2 + input_vector[1] ** 2 + input_vector[2] ** 2)
    return input_vector / (vector_norm + eps)


def get_perpendicular_vector(input_vector, eps=1e-15):
    """
    Returns a vector perpendicular to ``input_vector``

    :param input_vector: A 3d vector (:obj:`torch.tensor`)
    :param eps: Parameter used for numerical stability. Default is ``'1e-15'``
    """
    x1 = 1.
    y1 = 1.
    # x1 * input_vector[0] + y1 * input_vector[1] + z1 * input_vector[2] = 0
    z1 = - (x1 * input_vector[0] + y1 * input_vector[1]) / (input_vector[2] + eps)
    return vector3d(x1, y1, z1)



from gradoptics.optics.vector import vector3d, batch_vector, normalize_batch_vector, normalize_vector, dot_product

from gradoptics.distributions.gaussian_distribution import GaussianDistribution
from gradoptics.distributions.base_distribution import BaseDistribution
from gradoptics.distributions.atom_cloud import AtomCloud

from gradoptics.inference.rejection_sampling import rejection_sampling

from gradoptics.optics.lens import PerfectLens, ThickLens
from gradoptics.optics.mirror import FlatMirror, CurvedMirror
from gradoptics.optics.ray import Ray, Rays, empty_like, cat
from gradoptics.optics.sensor import Sensor
from gradoptics.optics.window import Window
from gradoptics.optics.camera import Camera
from gradoptics.optics.bounding_sphere import BoundingSphere

from gradoptics.ray_tracing.scene import Scene

from gradoptics.ray_tracing.ray_tracing import forward_ray_tracing
from gradoptics.ray_tracing.ray_tracing import backward_ray_tracing

from gradoptics.light_sources.light_source_from_distribution import LightSourceFromDistribution

from gradoptics.transforms import base_transform, look_at_transform, simple_transform

from gradoptics.optics.psf import PSF

from gradoptics.integrator.base_integrator import BaseIntegrator
from gradoptics.integrator.stratified_sampling_integrator import StratifiedSamplingIntegrator
